# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Vala Vakilian

from aiohttp import web, MultipartWriter, MultipartReader

from .routeutils import authenticate_by_token, authenticate_by_token_required_fields
from .routeutils import validate_required_fields, log_request, log


class UploadHandler:
    def __init__(self, plomServer):
        self.server = plomServer

    async def doesBundleExist(self, request):
        """Returns whether given bundle/md5sum known to database

        Checks both bundle's name and md5sum
        * neither = no matching bundle, return [False]
        * name but not md5 = return [True, 'name'] - user is trying to upload different bundles with same name.
        * md5 but not name = return [True, 'md5sum'] - user is trying to same bundle with different names.
        * both match = return [True, 'both'] - user could be retrying
          after network failure (for example) or uploading unknown or
          colliding pages.
        """
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "bundle", "md5sum"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] in ["scanner", "manager"]:
            return web.Response(status=401)
        rval = self.server.doesBundleExist(data["bundle"], data["md5sum"])
        return web.json_response(rval, status=200)  # all fine

    async def createNewBundle(self, request):
        """Try to create bundle with given name/md5sum.

        First check name / md5sum of bundle.
        * If bundle matches either 'name' or 'md5sum' then return [False, reason] - this shouldnt happen if scanner working correctly.
        * If bundle matches 'both' then return [True, skip_list] where skip_list = the page-orders from that bundle that are already in the system. The scan scripts will then skip those uploads.
        * If no such bundle return [True, []] - create the bundle and return an empty skip-list.

        Notes:
        * after declaring a bundle you may upload images to it.
        * uploading pages to an undeclared bundle is not allowed.
        * bundles traditionally correspond to one "pile" of physical
          papers scanned together.
        * there does not need to be one-to-one relationship betewen
          bundles and Exam Papers or Homework Papers.
        """
        log_request("createNewBundle", request)

        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "bundle", "md5sum"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] in ["scanner", "manager"]:
            return web.Response(status=401)
        rval = self.server.createNewBundle(data["bundle"], data["md5sum"])
        return web.json_response(rval, status=200)  # all fine

    async def sidToTest(self, request):
        """Match given student_id to a test-number.

        Returns
        * [True, test_number]
        * [False, 'Cannot find test with that student id']
        """
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "sid"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] in ["scanner", "manager"]:
            return web.Response(status=401)
        rval = self.server.sidToTest(data["sid"])
        return web.json_response(rval, status=200)

    async def uploadTestPage(self, request):
        """A test page has known page, known paper number, usually QR-coded.

        Typically the page is QR coded, and thus we know precisely what
        paper number, what question and what page.  We may not know the
        student depending on whether it was prenamed or not.

        Args:
            request (aiohttp.web_request.Request)

        Returns:
            aiohttp.web_response.Response: JSON data directly from the
                database call.

        Note: this uses the `status=200` success return code for some
        kinds of failures: it simply returns whatever data the DB gave
        back as blob of json for the client to deal with.  Thus, this
        API call is not recommended outside of Plom.
        """
        reader = MultipartReader.from_response(request)

        part0 = await reader.next()  # should be parameters
        if part0 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        param = await part0.json()

        if not validate_required_fields(
            param,
            [
                "user",
                "token",
                "test",
                "page",
                "version",
                "fileName",
                "md5sum",
                "bundle",
                "bundle_order",
            ],
        ):
            return web.Response(status=400)
        if not self.server.validate(param["user"], param["token"]):
            return web.Response(status=401)
        if not param["user"] in ("manager", "scanner"):
            return web.Response(status=401)

        # TODO: unused, we should ensure this matches the data
        # TODO: or why bother passing those in to param?
        code = request.match_info["tpv"]

        part1 = await reader.next()  # should be the image file
        if part1 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        image = await part1.read()
        # file it away.
        rmsg = self.server.addTestPage(
            param["test"],
            param["page"],
            param["version"],
            param["fileName"],
            image,
            param["md5sum"],
            param["bundle"],
            param["bundle_order"],
        )
        # note 200 used here for errors too
        return web.json_response(rmsg, status=200)

    async def uploadHWPage(self, request):
        """A homework page is self-scanned, known student, and known questions.

        Typically the page is without QR codes.  The uploader knows what
        student it belongs to and what question(s).  The order within the
        question is somewhat known too, at least within its upload bundle.

        Args:
            request (aiohttp.web_request.Request)

        Returns:
            aiohttp.web_response.Response: JSON data directly from the
                database call.

        The requests data has a `question` field, which can be a scalar
        or a list of questions we wish to upload too.  Maybe the scalar
        is deprecated?
        TODO: force it to always be a list?

        Note: this uses the `status=200` success return code for some
        kinds of failures: it simply returns whatever data the DB gave
        back as blob of json for the client to deal with.  Thus, this
        API call is not recommended outside of Plom.
        """
        reader = MultipartReader.from_response(request)

        part0 = await reader.next()  # should be parameters
        if part0 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        param = await part0.json()

        if not validate_required_fields(
            param,
            [
                "user",
                "token",
                "sid",
                "question",
                "order",
                "fileName",
                "md5sum",
                "bundle",
                "bundle_order",
            ],
        ):
            return web.Response(status=400)
        if not self.server.validate(param["user"], param["token"]):
            return web.Response(status=401)
        if not param["user"] in ("manager", "scanner"):
            return web.Response(status=401)

        part1 = await reader.next()  # should be the image file
        if part1 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        image = await part1.read()
        # file it away.
        rmsg = self.server.addHWPage(
            param["sid"],
            param["question"],
            param["order"],
            param["fileName"],
            image,
            param["md5sum"],
            param["bundle"],
            param["bundle_order"],
        )
        # note 200 used here for errors too
        return web.json_response(rmsg, status=200)

    async def uploadLPage(self, request):
        """A loose page is self-scanned, known student, but unknown question.

        Typically the page is without QR codes.  The uploader knows what
        student it belongs to but not what question.

        DEPRECATED? Perhaps on its way to deprecation if HW Pages become
        more general in the future.

        Args:
            request (aiohttp.web_request.Request)

        Returns:
            aiohttp.web_response.Response: JSON data directly from the
                database call.

        Note: this uses the `status=200` success return code for some
        kinds of failures: it simply returns whatever data the DB gave
        back as blob of json for the client to deal with.  Thus, this
        API call is not recommended outside of Plom.
        """
        reader = MultipartReader.from_response(request)

        part0 = await reader.next()  # should be parameters
        if part0 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        param = await part0.json()

        if not validate_required_fields(
            param,
            [
                "user",
                "token",
                "sid",
                "order",
                "fileName",
                "md5sum",
                "bundle",
                "bundle_order",
            ],
        ):
            return web.Response(status=400)
        if not self.server.validate(param["user"], param["token"]):
            return web.Response(status=401)
        if not param["user"] in ("manager", "scanner"):
            return web.Response(status=401)

        part1 = await reader.next()  # should be the image file
        if part1 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        image = await part1.read()
        # file it away.
        rmsg = self.server.addLPage(
            param["sid"],
            param["order"],
            param["fileName"],
            image,
            param["md5sum"],
            param["bundle"],
            param["bundle_order"],
        )
        # note 200 used here for errors too
        return web.json_response(rmsg, status=200)

    async def uploadUnknownPage(self, request):
        reader = MultipartReader.from_response(request)

        part0 = await reader.next()  # should be parameters
        if part0 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        param = await part0.json()

        if not validate_required_fields(
            param,
            ["user", "token", "fileName", "order", "md5sum", "bundle", "bundle_order"],
        ):
            return web.Response(status=400)
        if not self.server.validate(param["user"], param["token"]):
            return web.Response(status=401)
        if not param["user"] in ("manager", "scanner"):
            return web.Response(status=401)

        part1 = await reader.next()  # should be the image file
        if part1 is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        image = await part1.read()
        # file it away.
        rmsg = self.server.addUnknownPage(
            param["fileName"],
            image,
            param["order"],
            param["md5sum"],
            param["bundle"],
            param["bundle_order"],
        )
        # note 200 used here for errors too
        return web.json_response(rmsg, status=200)

    async def uploadCollidingPage(self, request):
        reader = MultipartReader.from_response(request)

        part0 = await reader.next()  # should be parameters
        if part0 is None:  # weird error
            return web.Response(status=406)  # should have sent 2 parts
        param = await part0.json()

        if not validate_required_fields(
            param,
            [
                "user",
                "token",
                "fileName",
                "md5sum",
                "test",
                "page",
                "version",
                "bundle",
                "bundle_order",
            ],
        ):
            return web.Response(status=400)
        if not self.server.validate(param["user"], param["token"]):
            return web.Response(status=401)
        # TODO - restrict to manager only.
        if not param["user"] in ("manager", "scanner"):
            return web.Response(status=401)

        # TODO: unused, we should ensure this matches the data
        code = request.match_info["tpv"]

        part1 = await reader.next()  # should be the image file
        if part1 is None:  # weird error
            return web.Response(status=406)  # should have sent 2 parts
        image = await part1.read()
        # file it away.
        rmsg = self.server.addCollidingPage(
            param["test"],
            param["page"],
            param["version"],
            param["fileName"],
            image,
            param["md5sum"],
            param["bundle"],
            param["bundle_order"],
        )
        # note 200 used here for errors too
        return web.json_response(rmsg, status=200)

    async def replaceMissingTestPage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "test", "page", "version"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.replaceMissingTestPage(
            data["test"], data["page"], data["version"]
        )
        if rval[0]:
            return web.json_response(rval, status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)  # page not found at all

    async def replaceMissingHWQuestion(self, request):
        # can replace either by SID-lookup or test-number
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "sid", "test", "question"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if data["user"] != "manager" and data["user"] != "scanner":
            return web.Response(status=401)

        rval = self.server.replaceMissingHWQuestion(
            data["sid"], data["test"], data["question"]
        )
        if rval[0]:
            return web.json_response(rval, status=200)  # all fine
        else:
            if rval[1] == "owners":
                return web.json_response(rval[2], status=409)
            elif rval[1] == "present":
                return web.Response(status=405)  # that question already has pages
            else:
                return web.Response(status=404)  # page not found at all

    async def removeAllScannedPages(self, request):
        data = await request.json()
        if not validate_required_fields(
            data,
            [
                "user",
                "token",
                "test",
            ],
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.removeAllScannedPages(
            data["test"],
        )
        if rval[0]:
            return web.json_response(rval, status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)  # page not found at all

    async def getUnknownPageNames(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getUnknownPageNames()
        return web.json_response(rval, status=200)  # all fine

    async def getDiscardNames(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getDiscardNames()
        return web.json_response(rval, status=200)  # all fine

    async def getCollidingPageNames(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getCollidingPageNames()
        return web.json_response(rval, status=200)  # all fine

    async def getTPageImage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "test", "page", "version"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getTPageImage(data["test"], data["page"], data["version"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getHWPageImage(self, request):  # should this use version too?
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "test", "question", "order"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getHWPageImage(data["test"], data["question"], data["order"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getEXPageImage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "test", "question", "order"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getEXPageImage(data["test"], data["question"], data["order"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getLPageImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "test", "order"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getLPageImage(data["test"], data["order"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getUnknownImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getUnknownImage(data["fileName"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getDiscardImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getDiscardImage(data["fileName"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    async def getCollidingImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.getCollidingImage(data["fileName"])
        if rval[0]:
            return web.FileResponse(rval[1], status=200)  # all fine
        else:
            return web.Response(status=404)

    # @route.get("/admin/questionImages")
    @authenticate_by_token_required_fields(["user", "test", "question"])
    def getQuestionImages(self, data, request):
        if not data["user"] == "manager":
            return web.Response(status=401)

        rmsg = self.server.getQuestionImages(data["test"], data["question"])
        # returns either [True, fname1,fname2,..,fname.n] or [False, error]
        if rmsg[0]:
            # insert number of parts [n, fn.1,fn.2,...fn.n]
            with MultipartWriter("images") as mpwriter:
                mpwriter.append(str(len(rmsg) - 1))
                for fn in rmsg[1:]:
                    mpwriter.append(open(fn, "rb"))
            return web.Response(body=mpwriter, status=200)
        else:
            return web.Response(status=404)  # couldnt find that test/question

    # @routes.get("/admin/testImages")
    @authenticate_by_token_required_fields(["user", "test"])
    def getAllTestImages(self, data, request):
        if not data["user"] == "manager":
            return web.Response(status=401)

        rmsg = self.server.getAllTestImages(data["test"])
        # returns either [True, fname1,fname2,..,fname.n] or [False, error]
        if rmsg[0]:
            # insert number of parts [n, fn.1,fn.2,...fn.n]
            with MultipartWriter("images") as mpwriter:
                mpwriter.append(str(len(rmsg) - 1))
                for fn in rmsg[1:]:
                    if fn == "":
                        mpwriter.append("")
                    else:
                        mpwriter.append(open(fn, "rb"))
            return web.Response(body=mpwriter, status=200)
        else:
            return web.Response(status=404)  # couldnt find that test/question

    async def checkTPage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "test", "page"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rmsg = self.server.checkTPage(data["test"], data["page"])
        # returns either [True, "collision", version, fname], [True, "scanned", version] or [False]
        if rmsg[0]:
            with MultipartWriter("images") as mpwriter:
                mpwriter.append("{}".format(rmsg[1]))  # append "collision" or "scanned"
                mpwriter.append("{}".format(rmsg[2]))  # append "version"
                if len(rmsg) == 4:  # append the image.
                    mpwriter.append(open(rmsg[3], "rb"))
            return web.Response(body=mpwriter, status=200)
        else:
            return web.Response(status=404)  # couldnt find that test/question

    async def removeUnknownImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.removeUnknownImage(data["fileName"])
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            return web.Response(status=404)

    async def removeCollidingImage(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.removeCollidingImage(data["fileName"])
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            return web.Response(status=404)

    async def unknownToTestPage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "fileName", "test", "page", "rotation"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.unknownToTestPage(
            data["fileName"], data["test"], data["page"], data["rotation"]
        )
        if rval[0]:
            return web.json_response(rval[1], status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)

    async def unknownToHWPage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "fileName", "test", "question", "rotation"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.unknownToHWPage(
            data["fileName"], data["test"], data["question"], data["rotation"]
        )
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)

    async def unknownToExtraPage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "fileName", "test", "question", "rotation"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.unknownToExtraPage(
            data["fileName"], data["test"], data["question"], data["rotation"]
        )  # returns [True], or [False, reason]
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)

    async def collidingToTestPage(self, request):
        data = await request.json()
        if not validate_required_fields(
            data, ["user", "token", "fileName", "test", "page", "version"]
        ):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.collidingToTestPage(
            data["fileName"], data["test"], data["page"], data["version"]
        )
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            if rval[1] == "owners":  # [False, "owners", owner_list]
                return web.json_response(rval[2], status=409)
            else:
                return web.Response(status=404)

    async def discardToUnknown(self, request):
        data = await request.json()
        if not validate_required_fields(data, ["user", "token", "fileName"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if not data["user"] == "manager":
            return web.Response(status=401)

        rval = self.server.discardToUnknown(data["fileName"])
        if rval[0]:
            return web.Response(status=200)  # all fine
        else:
            return web.Response(status=404)

    async def processHWUploads(self, request):
        """Trigger any updates that are appropriate after some uploads.

        This is probably similar to :py:meth:`processTUploads`
        """
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if data["user"] != "manager" and data["user"] != "scanner":
            return web.Response(status=401)

        update_count = self.server.processHWUploads()
        return web.json_response(update_count, status=200)

    async def processLUploads(self, request):
        """Trigger any updates that are appropriate after some uploads.

        This is probably similar to :py:meth:`processTUploads`
        """
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if data["user"] != "manager" and data["user"] != "scanner":
            return web.Response(status=401)

        update_count = self.server.processLUploads()
        return web.json_response(update_count, status=200)

    async def processTUploads(self, request):
        """Trigger any updates that are appropriate after some uploads.

        If we upload a bunch of pages to the server, the server will
        typically keep those in some sort of "staging" state where, for
        example, they are not given to marking clients.  This is b/c it
        will be distruptive to clients to have pages added to questions.
        To "release" a these recent uploads, we make this API call.

        Notes:
          * its ok to upload to a bundle after calling this (worse case,
            some client work will be invalidated or tagged to check).
          * its ok to call this repeatedly.
          * its not necessarily or useful to call this after uploading
            Unknown Pages or Colliding Pages: those will need to be
            dealt with in the Manager tool (e.g., added them to a
            Paper) at which time similar triggers will occur.

        Returns:
            aiohttp.web.Response: with status code as below.

        Status codes:
            200 OK: action was taken, report numer of Papers updated.
            401 Unauthorized: invalid credientials.
            403 Forbidden: only "manager"/"scanner" allowed to do this.
        """
        data = await request.json()
        if not validate_required_fields(data, ["user", "token"]):
            return web.Response(status=400)
        if not self.server.validate(data["user"], data["token"]):
            return web.Response(status=401)
        if data["user"] != "manager" and data["user"] != "scanner":
            return web.Response(status=403)

        update_count = self.server.processTUploads()
        return web.json_response(update_count, status=200)

    @authenticate_by_token_required_fields(["user"])
    def populateExamDatabase(self, data, request):
        """Instruct the server to generate paper data in the database.

        TODO: maybe the api call should just be for one row of the database.

        TODO: or maybe we can pass the page-to-version mapping to this?
        """
        if not data["user"] == "manager":
            return web.Response(status=400)  # malformed request.

        from plom.db import buildExamDatabaseFromSpec

        # TODO this is not the design we have elsewhere, should call helper function
        try:
            r, summary = buildExamDatabaseFromSpec(self.server.testSpec, self.server.DB)
        except ValueError:
            raise web.HTTPConflict(
                reason="Database already present: not overwriting"
            ) from None
        if r:
            return web.Response(text=summary, status=200)
        else:
            raise web.HTTPInternalServerError(text=summary)

    # TODO: would be nice to use @authenticate_by_token, see comments in routeutils.py
    @authenticate_by_token_required_fields([])
    def getPageVersionMap(self, data, request):
        """Get the mapping between page number and version for one test.

        Returns:
            dict: keyed by page number.  Note keys are strings b/c of
                json limitations; you may want to convert back to int.
        """
        spec = self.server.testSpec
        paper_idx = request.match_info["papernum"]
        ver = self.server.DB.getPageVersions(paper_idx)
        if ver:
            return web.json_response(ver, status=200)
        else:
            return web.Response(status=404)

    @authenticate_by_token_required_fields([])
    def getGlobalPageVersionMap(self, data, request):
        """Get the mapping between page number and version for all tests.

        Returns:
            dict: dict of dicts, keyed first by paper index then by page
                number.  Both keys are strings b/c of json limitations;
                you may need to iterate and convert back to int.  Fails
                with 500 Internal Server Error if a test does not exist.
        """
        spec = self.server.testSpec
        vers = {}
        for paper_idx in range(1, spec["numberToProduce"] + 1):
            ver = self.server.DB.getPageVersions(paper_idx)
            if not ver:
                return web.Response(status=500)
            vers[paper_idx] = ver
        # JSON converts int keys to strings, we'll fix this at the far end
        # return web.json_response(str(pickle.dumps(vers)), status=200)
        return web.json_response(vers, status=200)

    # @route.put("/admin/pdf_produced/{t}")
    @authenticate_by_token_required_fields(["user"])
    def notify_pdf_of_paper_produced(self, data, request):
        """Inform server that a PDF for this paper has been produced.

        This is to be called one-at-a-time for each paper.  If this is a
        bottleneck we could consider adding a "bulk" version.

        Note that the file itself is not uploaded to the server: we're
        just merely creating a record that such a file exists somewhere.

        TODO: pass in md5sum too and if its unchanged no need to
        complain about conflict, just quietly return 200.
        TODO: implement force as mentioned below.

        Inputs:
            t (int?, str?): part of URL that specifies the paper number.
            user (str): who's calling?  A field of the request.
            force (bool): force production even if paper already exists.
            md5sum (str): md5sum of the file that was produced.

        Returns:
            aiohttp.web.Response: with status code as below.

        Status codes:
            200 OK: the info was recorded.
            400 Bad Request: only "manager" is allowed to do this.
            401 Unauthorized: invalid credientials.
            404 Not Found: paper number is outside valid range.
            409 Conflict: this paper has already been produced, so its
                unusual to be making it again. Maybe try `force=True`.
        """
        if not data["user"] == "manager":
            return web.Response(status=400)
        # force_flag = request.match_info["force"]
        paper_idx = request.match_info["papernum"]
        try:
            self.server.DB.produceTest(paper_idx)
        except IndexError:
            return web.Response(status=404)
        except ValueError:
            return web.Response(status=409)
        return web.Response(status=200)

    def setUpRoutes(self, router):
        router.add_get("/admin/bundle", self.doesBundleExist)
        router.add_put("/admin/bundle", self.createNewBundle)
        router.add_get("/admin/sidToTest", self.sidToTest)
        router.add_put("/admin/testPages/{tpv}", self.uploadTestPage)
        router.add_put("/admin/hwPages", self.uploadHWPage)
        router.add_put("/admin/lPages", self.uploadLPage)
        router.add_put("/admin/unknownPages", self.uploadUnknownPage)
        router.add_put("/admin/collidingPages/{tpv}", self.uploadCollidingPage)
        router.add_put("/admin/missingTestPage", self.replaceMissingTestPage)
        router.add_put("/admin/missingHWQuestion", self.replaceMissingHWQuestion)
        router.add_delete("/admin/scannedPages", self.removeAllScannedPages)
        router.add_get("/admin/scannedTPage", self.getTPageImage)
        router.add_get("/admin/scannedHWPage", self.getHWPageImage)
        router.add_get("/admin/scannedEXPage", self.getEXPageImage)
        router.add_get("/admin/scannedLPage", self.getLPageImage)
        router.add_get("/admin/unknownPageNames", self.getUnknownPageNames)
        router.add_get("/admin/discardNames", self.getDiscardNames)
        router.add_get("/admin/collidingPageNames", self.getCollidingPageNames)
        router.add_get("/admin/unknownImage", self.getUnknownImage)
        router.add_get("/admin/discardImage", self.getDiscardImage)
        router.add_get("/admin/collidingImage", self.getCollidingImage)
        router.add_get("/admin/questionImages", self.getQuestionImages)
        router.add_get("/admin/testImages", self.getAllTestImages)
        router.add_get("/admin/checkTPage", self.checkTPage)
        router.add_delete("/admin/unknownImage", self.removeUnknownImage)
        router.add_delete("/admin/collidingImage", self.removeCollidingImage)
        router.add_put("/admin/unknownToTestPage", self.unknownToTestPage)
        router.add_put("/admin/unknownToHWPage", self.unknownToHWPage)
        router.add_put("/admin/unknownToExtraPage", self.unknownToExtraPage)
        router.add_put("/admin/collidingToTestPage", self.collidingToTestPage)
        router.add_put("/admin/discardToUnknown", self.discardToUnknown)
        router.add_put("/admin/hwPagesUploaded", self.processHWUploads)
        router.add_put("/admin/loosePagesUploaded", self.processLUploads)
        router.add_put("/admin/testPagesUploaded", self.processTUploads)
        router.add_put("/admin/populateDB", self.populateExamDatabase)
        router.add_get("/admin/pageVersionMap/{papernum}", self.getPageVersionMap)
        router.add_get("/admin/pageVersionMap", self.getGlobalPageVersionMap)
        router.add_put(
            "/admin/pdf_produced/{papernum}", self.notify_pdf_of_paper_produced
        )



#!/usr/bin/env python3

# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020-2021 Colin B. Macdonald
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Plom tools for scanning tests and pushing to servers.

## Overview of the scanning process

  1. Decide on a working directory for your scans, copy your PDFs into
     that directory and then cd into it.

  2. Use the `process` command to split your first PDF into bitmaps of
     each page.  This will also read any QR codes from the pages and
     match these against expectations from the server.

  3. Use the `upload` command to send pages to the server.  There are
     additional flags for dealing with special cases:

       a. Pages that could not be identified are called "Unknowns".
          They can include "Extra Pages" without QR codes, poor-quality
          scans where the QR reader failed, folded papers, etc.  A small
          number is normal but large numbers are cause for concern and
          sanity checking.  A human will (eventually) have to identify
          these manually.

       b. If the system detects you trying to upload a test page
          corresponding to one already in the system (but not identical)
          then those pages are filed as "Collisions". If you have good
          paper-handling protocols then this should not happen, except
          in exceptional circumstances (such as rescanning an illegible
          page).  Force the upload these if you really need to; the
          manager will then have to look at them.

  4. Run "plom-scan status" to get a brief summary of scanning to date.

  5. If something goes wrong such as crashes or interruptions, you may
     need to clear the "scanner" login with the `clear` command.

  These steps may be repeated as new PDF files come in: it is not
  necessary to wait until scanning is complete to start processing and
  uploading.
"""

__copyright__ = "Copyright (C) 2020-2021 Andrew Rechnitzer, Colin B. Macdonald et al"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import argparse
import os
from pathlib import Path
from warnings import warn

import toml

from plom import __version__
from plom.scan import (
    upload_unknowns,
    print_unknowns_warning,
    bundle_has_nonuploaded_unknowns,
    upload_collisions,
    print_collision_warning,
    bundle_has_nonuploaded_collisions,
    bundle_name_and_md5,
)


# TODO: this bit of code from messenger could be useful here
#    if os.path.isfile("server.toml"):
#        with open("server.toml") as fh:
#            si = toml.load(fh)
#        server = si["server"]
#        if server and ":" in server:
#            server, message_port = server.split(":")


# TODO: make some common util file to store all these names?
archivedir = Path("archivedPDFs")


def clearLogin(server, password):
    from plom.scan import clearScannerLogin

    clearScannerLogin.clearLogin(server, password)


def scanStatus(server, password):
    from plom.scan import checkScanStatus

    checkScanStatus.checkStatus(server, password)


def make_required_directories(bundle=None):
    os.makedirs(archivedir, exist_ok=True)
    os.makedirs("bundles", exist_ok=True)
    # TODO: split up a bit, above are global, below per bundle
    if bundle:
        directory_list = [
            "uploads/sentPages",
            "uploads/discardedPages",
            "uploads/collidingPages",
            "uploads/sentPages/unknowns",
            "uploads/sentPages/collisions",
        ]
        for dir in directory_list:
            os.makedirs(bundle / Path(dir), exist_ok=True)


def processScans(server, password, pdf_fname, gamma, extractbmp):
    """Process PDF file into images and read QRcodes

    Convert file into a bundle-name
    Check with server if bundle/md5 already on server
    - abort if name xor md5sum known.
    - continue if neither known or both known
    Make required directories for processing bundle,
    convert PDF to images and read QR codes from those.
    """
    from plom.scan import scansToImages
    from plom.scan import sendPagesToServer
    from plom.scan import readQRCodes

    pdf_fname = Path(pdf_fname)
    if not pdf_fname.is_file():
        print("Cannot find file {} - skipping".format(pdf_fname))
        return
    # TODO: replace above with letting exception rise from next:
    bundle_name, md5 = bundle_name_and_md5(pdf_fname)
    # TODO: doesBundleExist(bundle_name, md5)

    print("Checking if bundle {} already exists on server".format(bundle_name))
    bundle_exists = sendPagesToServer.doesBundleExist(pdf_fname, server, password)
    # return [False, name], [True, name], [True,md5sum] or [True, both]
    if bundle_exists[0]:
        if bundle_exists[1] == "name":
            print(
                "The bundle name {} has been used previously for a different bundle. Stopping".format(
                    pdf_fname
                )
            )
            return
        elif bundle_exists[1] == "md5sum":
            print(
                "A bundle with matching md5sum is already in system with a different name. Stopping"
            )
            return
        elif bundle_exists[1] == "both":
            print(
                "Warning - bundle {} has been declared previously - you are likely trying again as a result of a crash. Continuing".format(
                    bundle_name
                )
            )
        else:
            raise RuntimeError("Should not be here: unexpected code path!")

    bundledir = Path("bundles") / bundle_name
    make_required_directories(bundledir)

    with open(bundledir / "source.toml", "w+") as f:
        toml.dump({"file": str(pdf_fname), "md5": md5}, f)

    print("Processing PDF {} to images".format(pdf_fname))
    scansToImages.processScans(pdf_fname, bundledir, not gamma, not extractbmp)
    print("Read QR codes")
    readQRCodes.processBitmaps(bundledir, server, password)
    # TODO: can collisions warning be written here too?
    if bundle_has_nonuploaded_unknowns(bundledir):
        print_unknowns_warning(bundledir)
        print('You can upload these by passing "--unknowns" to the upload command')


def uploadImages(
    server, password, bundle_name, unknowns_flag=False, collisions_flag=False
):
    """Upload processed images from bundle.

    Try to create a bundle on server.
    - abort if name xor md5sum of bundle known.
    - continue otherwise (server will give skip-list)
    Skip images whose page within the bundle is in the skip-list since
    those are already uploaded.
    Once uploaded archive the bundle pdf.

    As part of the upload 'unknown' pages and 'collisions' may be detected.
    These will not be uploaded unless the appropriate flags are set.
    """
    from plom.scan import sendPagesToServer, scansToImages

    if bundle_name.lower().endswith(".pdf"):
        warn('Careful, the bundle name should not include ".pdf"')

    bundledir = Path("bundles") / bundle_name
    info = toml.load(bundledir / "source.toml")
    md5 = info["md5"]

    # TODO: check first to avoid misleading msg?
    print('Creating bundle "{}" on server'.format(bundle_name))
    rval = sendPagesToServer.createNewBundle(bundle_name, md5, server, password)
    # should be [True, skip_list] or [False, reason]
    if rval[0]:
        skip_list = rval[1]
        if len(skip_list) > 0:
            print("Some images from that bundle were uploaded previously:")
            print("Pages {}".format(skip_list))
            print("Skipping those images.")
    else:
        print("There was a problem with this bundle.")
        if rval[1] == "name":
            print("A different bundle with the same name was uploaded previously.")
        else:
            print(
                "A bundle with matching md5sum but different name was uploaded previously."
            )
        print("Stopping.")
        return

    print("Upload images to server")
    [TPN, updates] = sendPagesToServer.uploadTPages(
        bundledir, skip_list, server, password
    )
    print(
        "Tests were uploaded to the following studentIDs: {}".format(
            ", ".join(TPN.keys())
        )
    )
    print("Server reports {} papers updated.".format(updates))

    pdf_fname = Path(info["file"])
    if pdf_fname.exists():
        print(
            'Original PDF "{}" still in place: archiving to "{}"...'.format(
                pdf_fname, str(archivedir)
            )
        )
        scansToImages.archiveTBundle(pdf_fname)
    elif (archivedir / pdf_fname).exists():
        print(
            'Original PDF "{}" is already archived in "{}".'.format(
                pdf_fname, str(archivedir)
            )
        )
    else:
        raise RuntimeError("Did you move the archived PDF?  Please don't do that!")

    # Note: no need to "finalize" a bundle, its ok to send unknown/collisions
    # after the above call to sendPagesToServer.

    if unknowns_flag:
        if bundle_has_nonuploaded_unknowns(bundledir):
            print_unknowns_warning(bundledir)
            print("Unknowns upload flag present: uploading...")
            upload_unknowns(bundledir, server, password)
        else:
            print(
                "Unknowns upload flag present: but no unknowns - so no actions required."
            )
    else:
        if bundle_has_nonuploaded_unknowns(bundledir):
            print_unknowns_warning(bundledir)
            print('If you want to upload these unknowns, rerun with "--unknowns".')

    if collisions_flag:
        if bundle_has_nonuploaded_collisions(bundledir):
            print_collision_warning(bundledir)
            print("Collisions upload flag present.")
            # TODO:add a --yes flag?
            yn = input("Are you sure you want to upload these colliding pages? [y/N] ")
            if yn.lower() == "y":
                print("Proceeding.")
                upload_collisions(bundledir, server, password)
        else:
            print(
                "Collisions upload flag present: but no collisions - so no actions required."
            )
    else:
        if bundle_has_nonuploaded_collisions(bundledir):
            print_collision_warning(bundledir)
            print('If you want to upload these collisions, rerun with "--collisions".')


parser = argparse.ArgumentParser(
    description=__doc__.split("\n")[0],
    epilog="\n".join(__doc__.split("\n")[1:]),
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)
sub = parser.add_subparsers(dest="command")

spP = sub.add_parser(
    "process",
    help="Process scanned PDF to images and read QRs",
    description="Process one scanned PDF into page images, read QR codes and check info with server (e.g., versions match).",
)
spU = sub.add_parser(
    "upload",
    help="Upload page images to scanner",
    description="Upload page images to scanner.",
)
spS = sub.add_parser(
    "status",
    help="Get scanning status report from server",
    description="Get scanning status report from server.",
)
spC = sub.add_parser(
    "clear",
    help='Clear "scanner" login',
    description='Clear "scanner" login after a crash or other expected event.',
)
# TODO: maybe in the future?
# spA = sub.add_parser(
#     "all",
#     help="Process, read and upload page images to scanner (WIP!)",
#     description="Process, read and upload page images to scanner. CAUTION: Work in Progress!",
# )
# spA.add_argument("scanPDF", nargs="+", help="The PDF(s) containing scanned pages.")
spP.add_argument("scanPDF", help="The PDF file of scanned pages.")
g = spP.add_mutually_exclusive_group(required=False)
g.add_argument(
    "--gamma-shift",
    action="store_true",
    dest="gamma",
    help="""
        Apply white balancing to the scan, if the image format is
        lossless (PNG).
        By default, this gamma shift is NOT applied; this is because it
        may worsen some poor-quality scans with large shadow regions.
    """,
)
g.add_argument(
    "--no-gamma-shift",
    action="store_false",
    dest="gamma",
    help="Do not apply white balancing.",
)
g = spP.add_mutually_exclusive_group(required=False)
g.add_argument(
    "--extract-bitmaps",
    action="store_true",
    dest="extractbmp",
    help="""
        If a PDF page seems to contain exactly one bitmap image and
        nothing else, then extract that losslessly instead of rendering
        the page as a new PNG file.  This will typically give nicer
        images for the common scan case where pages are simply JPEG
        images.  But some care must be taken that the image is not
        annotated in any way and that no other markings appear on the
        page.
        As the algorithm to decide this is NOT YET IDEAL, this is
        currently OFF BY DEFAULT, but we anticipate it being the default
        in a future version.
    """,
)
g.add_argument(
    "--no-extract-bitmaps",
    action="store_false",
    dest="extractbmp",
    help="""
        Don't try to extract bitmaps; just render each page.  This is
        safer but not always ideal for image quality.
    """,
)

spU.add_argument("bundleName", help="The name of the PDF file, without extension.")
spU.add_argument(
    "-u",
    "--unknowns",
    action="store_true",
    help='Upload "unknowns", pages from which the QR-codes could not be read.',
)
spU.add_argument(
    "-c",
    "--collisions",
    action="store_true",
    help='Upload "collisions", pages which appear to already be on the server. '
    + "You should not need this option except under exceptional circumstances.",
)
for x in (spU, spS, spC, spP):
    x.add_argument("-s", "--server", metavar="SERVER[:PORT]", action="store")
    x.add_argument("-w", "--password", type=str, help='for the "scanner" user')


def main():
    args = parser.parse_args()

    if not hasattr(args, "server") or not args.server:
        try:
            args.server = os.environ["PLOM_SERVER"]
        except KeyError:
            pass
    if not hasattr(args, "password") or not args.password:
        try:
            args.password = os.environ["PLOM_SCAN_PASSWORD"]
        except KeyError:
            pass

    if args.command == "process":
        processScans(
            args.server, args.password, args.scanPDF, args.gamma, args.extractbmp
        )
    elif args.command == "upload":
        uploadImages(
            args.server, args.password, args.bundleName, args.unknowns, args.collisions
        )
    elif args.command == "status":
        scanStatus(args.server, args.password)
    elif args.command == "clear":
        clearLogin(args.server, args.password)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qtCreatorFiles/ui_annotator_rhm.ui'
#
# Created by: PyQt5 UI code generator 5.14.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_annotator_rhm(object):
    def setupUi(self, annotator_rhm):
        annotator_rhm.setObjectName("annotator_rhm")
        annotator_rhm.setWindowModality(QtCore.Qt.WindowModal)
        annotator_rhm.resize(862, 670)
        self.horizontalLayout = QtWidgets.QHBoxLayout(annotator_rhm)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.hideableBox = QtWidgets.QFrame(annotator_rhm)
        self.hideableBox.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.hideableBox.setFrameShadow(QtWidgets.QFrame.Raised)
        self.hideableBox.setObjectName("hideableBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.hideableBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame1 = QtWidgets.QFrame(self.hideableBox)
        self.frame1.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame1.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame1.setObjectName("frame1")
        self.modeLayout = QtWidgets.QHBoxLayout(self.frame1)
        self.modeLayout.setContentsMargins(0, 0, 0, 0)
        self.modeLayout.setObjectName("modeLayout")
        self.hamMenuButton = QtWidgets.QToolButton(self.frame1)
        self.hamMenuButton.setText("☰")
        self.hamMenuButton.setObjectName("hamMenuButton")
        self.modeLayout.addWidget(self.hamMenuButton)
        self.finishedButton = QtWidgets.QPushButton(self.frame1)
        self.finishedButton.setWhatsThis("")
        self.finishedButton.setObjectName("finishedButton")
        self.modeLayout.addWidget(self.finishedButton)
        self.finishNoRelaunchButton = QtWidgets.QPushButton(self.frame1)
        self.finishNoRelaunchButton.setObjectName("finishNoRelaunchButton")
        self.modeLayout.addWidget(self.finishNoRelaunchButton)
        self.cancelButton = QtWidgets.QPushButton(self.frame1)
        self.cancelButton.setObjectName("cancelButton")
        self.modeLayout.addWidget(self.cancelButton)
        self.verticalLayout.addWidget(self.frame1)
        self.frame2 = QtWidgets.QFrame(self.hideableBox)
        self.frame2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame2.setObjectName("frame2")
        self.buttonsLayout = QtWidgets.QHBoxLayout(self.frame2)
        self.buttonsLayout.setContentsMargins(0, 0, 0, 0)
        self.buttonsLayout.setSpacing(3)
        self.buttonsLayout.setObjectName("buttonsLayout")
        self.markLabel = QtWidgets.QLabel(self.frame2)
        self.markLabel.setObjectName("markLabel")
        self.buttonsLayout.addWidget(self.markLabel)
        self.zoomCB = QtWidgets.QComboBox(self.frame2)
        self.zoomCB.setObjectName("zoomCB")
        self.buttonsLayout.addWidget(self.zoomCB)
        self.verticalLayout.addWidget(self.frame2)
        self.frameTools = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frameTools.sizePolicy().hasHeightForWidth())
        self.frameTools.setSizePolicy(sizePolicy)
        self.frameTools.setObjectName("frameTools")
        self.toolLayout = QtWidgets.QGridLayout(self.frameTools)
        self.toolLayout.setContentsMargins(0, 0, 0, 0)
        self.toolLayout.setSpacing(3)
        self.toolLayout.setObjectName("toolLayout")
        self.penButton = QtWidgets.QToolButton(self.frameTools)
        self.penButton.setMinimumSize(QtCore.QSize(45, 0))
        self.penButton.setCheckable(True)
        self.penButton.setAutoExclusive(True)
        self.penButton.setObjectName("penButton")
        self.toolLayout.addWidget(self.penButton, 0, 4, 1, 1)
        self.textButton = QtWidgets.QToolButton(self.frameTools)
        self.textButton.setMinimumSize(QtCore.QSize(45, 0))
        self.textButton.setCheckable(True)
        self.textButton.setAutoExclusive(True)
        self.textButton.setObjectName("textButton")
        self.toolLayout.addWidget(self.textButton, 1, 4, 1, 1)
        self.commentDownButton = QtWidgets.QToolButton(self.frameTools)
        self.commentDownButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentDownButton.setToolTipDuration(-1)
        self.commentDownButton.setObjectName("commentDownButton")
        self.toolLayout.addWidget(self.commentDownButton, 2, 3, 1, 1)
        self.commentUpButton = QtWidgets.QToolButton(self.frameTools)
        self.commentUpButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentUpButton.setToolTipDuration(-1)
        self.commentUpButton.setObjectName("commentUpButton")
        self.toolLayout.addWidget(self.commentUpButton, 0, 3, 1, 1)
        self.panButton = QtWidgets.QToolButton(self.frameTools)
        self.panButton.setMinimumSize(QtCore.QSize(45, 0))
        self.panButton.setCheckable(True)
        self.panButton.setAutoExclusive(True)
        self.panButton.setObjectName("panButton")
        self.toolLayout.addWidget(self.panButton, 0, 0, 1, 1)
        self.lineButton = QtWidgets.QToolButton(self.frameTools)
        self.lineButton.setMinimumSize(QtCore.QSize(45, 0))
        self.lineButton.setCheckable(True)
        self.lineButton.setAutoExclusive(True)
        self.lineButton.setObjectName("lineButton")
        self.toolLayout.addWidget(self.lineButton, 2, 4, 1, 1)
        self.tickButton = QtWidgets.QToolButton(self.frameTools)
        self.tickButton.setMinimumSize(QtCore.QSize(45, 0))
        self.tickButton.setToolTipDuration(-1)
        self.tickButton.setCheckable(True)
        self.tickButton.setAutoExclusive(True)
        self.tickButton.setObjectName("tickButton")
        self.toolLayout.addWidget(self.tickButton, 1, 2, 1, 1)
        self.moveButton = QtWidgets.QToolButton(self.frameTools)
        self.moveButton.setMinimumSize(QtCore.QSize(45, 0))
        self.moveButton.setCheckable(True)
        self.moveButton.setAutoExclusive(True)
        self.moveButton.setObjectName("moveButton")
        self.toolLayout.addWidget(self.moveButton, 2, 0, 1, 1)
        self.redoButton = QtWidgets.QToolButton(self.frameTools)
        self.redoButton.setMinimumSize(QtCore.QSize(45, 0))
        self.redoButton.setObjectName("redoButton")
        self.toolLayout.addWidget(self.redoButton, 0, 1, 1, 1)
        self.zoomButton = QtWidgets.QToolButton(self.frameTools)
        self.zoomButton.setMinimumSize(QtCore.QSize(45, 0))
        self.zoomButton.setCheckable(True)
        self.zoomButton.setAutoExclusive(True)
        self.zoomButton.setObjectName("zoomButton")
        self.toolLayout.addWidget(self.zoomButton, 1, 0, 1, 1)
        self.deleteButton = QtWidgets.QToolButton(self.frameTools)
        self.deleteButton.setMinimumSize(QtCore.QSize(45, 0))
        self.deleteButton.setCheckable(True)
        self.deleteButton.setAutoExclusive(True)
        self.deleteButton.setObjectName("deleteButton")
        self.toolLayout.addWidget(self.deleteButton, 2, 1, 1, 1)
        self.commentButton = QtWidgets.QToolButton(self.frameTools)
        self.commentButton.setMinimumSize(QtCore.QSize(45, 0))
        self.commentButton.setToolTipDuration(-1)
        self.commentButton.setCheckable(True)
        self.commentButton.setAutoExclusive(True)
        self.commentButton.setObjectName("commentButton")
        self.toolLayout.addWidget(self.commentButton, 1, 3, 1, 1)
        self.undoButton = QtWidgets.QToolButton(self.frameTools)
        self.undoButton.setMinimumSize(QtCore.QSize(45, 0))
        self.undoButton.setObjectName("undoButton")
        self.toolLayout.addWidget(self.undoButton, 1, 1, 1, 1)
        self.crossButton = QtWidgets.QToolButton(self.frameTools)
        self.crossButton.setMinimumSize(QtCore.QSize(45, 0))
        self.crossButton.setCheckable(True)
        self.crossButton.setAutoExclusive(True)
        self.crossButton.setObjectName("crossButton")
        self.toolLayout.addWidget(self.crossButton, 0, 2, 1, 1)
        self.boxButton = QtWidgets.QToolButton(self.frameTools)
        self.boxButton.setMinimumSize(QtCore.QSize(45, 0))
        self.boxButton.setCheckable(True)
        self.boxButton.setAutoExclusive(True)
        self.boxButton.setObjectName("boxButton")
        self.toolLayout.addWidget(self.boxButton, 2, 2, 1, 1)
        self.verticalLayout.addWidget(self.frameTools, 0, QtCore.Qt.AlignHCenter)
        self.frame4 = QtWidgets.QFrame(self.hideableBox)
        self.frame4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame4.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame4.setObjectName("frame4")
        self.ebLayout = QtWidgets.QHBoxLayout(self.frame4)
        self.ebLayout.setContentsMargins(0, 0, 0, 0)
        self.ebLayout.setSpacing(3)
        self.ebLayout.setObjectName("ebLayout")
        self.modeLabel = QtWidgets.QLabel(self.frame4)
        self.modeLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.modeLabel.setObjectName("modeLabel")
        self.ebLayout.addWidget(self.modeLabel)
        self.verticalLayout.addWidget(self.frame4)
        self.frame5_markgrid = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame5_markgrid.sizePolicy().hasHeightForWidth())
        self.frame5_markgrid.setSizePolicy(sizePolicy)
        self.frame5_markgrid.setObjectName("frame5_markgrid")
        self.container_markgrid = QtWidgets.QHBoxLayout(self.frame5_markgrid)
        self.container_markgrid.setContentsMargins(0, -1, 0, -1)
        self.container_markgrid.setObjectName("container_markgrid")
        self.verticalLayout.addWidget(self.frame5_markgrid)
        self.frame6 = QtWidgets.QFrame(self.hideableBox)
        self.frame6.setObjectName("frame6")
        self.hboxlayout = QtWidgets.QHBoxLayout(self.frame6)
        self.hboxlayout.setContentsMargins(0, 0, 0, 0)
        self.hboxlayout.setObjectName("hboxlayout")
        self.noAnswerButton = QtWidgets.QPushButton(self.frame6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.noAnswerButton.sizePolicy().hasHeightForWidth())
        self.noAnswerButton.setSizePolicy(sizePolicy)
        self.noAnswerButton.setObjectName("noAnswerButton")
        self.hboxlayout.addWidget(self.noAnswerButton)
        spacerItem = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.hboxlayout.addItem(spacerItem)
        self.rearrangePagesButton = QtWidgets.QPushButton(self.frame6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rearrangePagesButton.sizePolicy().hasHeightForWidth())
        self.rearrangePagesButton.setSizePolicy(sizePolicy)
        self.rearrangePagesButton.setObjectName("rearrangePagesButton")
        self.hboxlayout.addWidget(self.rearrangePagesButton)
        self.verticalLayout.addWidget(self.frame6)
        self.frame7_comment = QtWidgets.QFrame(self.hideableBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(2)
        sizePolicy.setHeightForWidth(self.frame7_comment.sizePolicy().hasHeightForWidth())
        self.frame7_comment.setSizePolicy(sizePolicy)
        self.frame7_comment.setObjectName("frame7_comment")
        self.container_commentwidget = QtWidgets.QVBoxLayout(self.frame7_comment)
        self.container_commentwidget.setContentsMargins(0, -1, 0, 0)
        self.container_commentwidget.setObjectName("container_commentwidget")
        self.verticalLayout.addWidget(self.frame7_comment)
        self.horizontalLayout.addWidget(self.hideableBox)
        self.revealBox0 = QtWidgets.QFrame(annotator_rhm)
        self.revealBox0.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox0.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox0.setObjectName("revealBox0")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.revealBox0)
        self.verticalLayout_2.setSpacing(2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.revealBox1 = QtWidgets.QFrame(self.revealBox0)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.revealBox1.sizePolicy().hasHeightForWidth())
        self.revealBox1.setSizePolicy(sizePolicy)
        self.revealBox1.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox1.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox1.setObjectName("revealBox1")
        self.revealLayout = QtWidgets.QGridLayout(self.revealBox1)
        self.revealLayout.setContentsMargins(0, 0, 0, 0)
        self.revealLayout.setSpacing(3)
        self.revealLayout.setObjectName("revealLayout")
        self.deltaButton = QtWidgets.QToolButton(self.revealBox1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.deltaButton.sizePolicy().hasHeightForWidth())
        self.deltaButton.setSizePolicy(sizePolicy)
        self.deltaButton.setMinimumSize(QtCore.QSize(45, 0))
        self.deltaButton.setCheckable(True)
        self.deltaButton.setAutoExclusive(True)
        self.deltaButton.setObjectName("deltaButton")
        self.revealLayout.addWidget(self.deltaButton, 1, 0, 1, 1)
        self.verticalLayout_2.addWidget(self.revealBox1, 0, QtCore.Qt.AlignTop)
        self.revealBox2 = QtWidgets.QFrame(self.revealBox0)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.revealBox2.sizePolicy().hasHeightForWidth())
        self.revealBox2.setSizePolicy(sizePolicy)
        self.revealBox2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.revealBox2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.revealBox2.setObjectName("revealBox2")
        self.revealLayout2 = QtWidgets.QVBoxLayout(self.revealBox2)
        self.revealLayout2.setContentsMargins(3, 3, 3, 3)
        self.revealLayout2.setSpacing(3)
        self.revealLayout2.setObjectName("revealLayout2")
        self.verticalLayout_2.addWidget(self.revealBox2, 0, QtCore.Qt.AlignHCenter|QtCore.Qt.AlignBottom)
        self.horizontalLayout.addWidget(self.revealBox0)
        self.pageFrame = QtWidgets.QFrame(annotator_rhm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pageFrame.sizePolicy().hasHeightForWidth())
        self.pageFrame.setSizePolicy(sizePolicy)
        self.pageFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.pageFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.pageFrame.setObjectName("pageFrame")
        self.pageFrameGrid = QtWidgets.QGridLayout(self.pageFrame)
        self.pageFrameGrid.setContentsMargins(0, 0, 0, 0)
        self.pageFrameGrid.setObjectName("pageFrameGrid")
        self.horizontalLayout.addWidget(self.pageFrame)

        self.retranslateUi(annotator_rhm)
        QtCore.QMetaObject.connectSlotsByName(annotator_rhm)

    def retranslateUi(self, annotator_rhm):
        _translate = QtCore.QCoreApplication.translate
        annotator_rhm.setWindowTitle(_translate("annotator_rhm", "Annotate paper"))
        self.finishedButton.setToolTip(_translate("annotator_rhm", "Save and move to the next paper"))
        self.finishedButton.setText(_translate("annotator_rhm", "Next"))
        self.finishNoRelaunchButton.setToolTip(_translate("annotator_rhm", "Save and return to marker window"))
        self.finishNoRelaunchButton.setText(_translate("annotator_rhm", "Done"))
        self.cancelButton.setToolTip(_translate("annotator_rhm", "Cancel the current annotations and return to marker window"))
        self.cancelButton.setText(_translate("annotator_rhm", "&Cancel"))
        self.markLabel.setText(_translate("annotator_rhm", "kk out of nn"))
        self.penButton.setToolTip(_translate("annotator_rhm", "press t"))
        self.penButton.setText(_translate("annotator_rhm", "..."))
        self.textButton.setToolTip(_translate("annotator_rhm", "press g"))
        self.textButton.setText(_translate("annotator_rhm", "..."))
        self.commentDownButton.setToolTip(_translate("annotator_rhm", "press v"))
        self.commentDownButton.setText(_translate("annotator_rhm", "..."))
        self.commentUpButton.setToolTip(_translate("annotator_rhm", "press r"))
        self.commentUpButton.setText(_translate("annotator_rhm", "..."))
        self.panButton.setToolTip(_translate("annotator_rhm", "press q"))
        self.panButton.setText(_translate("annotator_rhm", "..."))
        self.lineButton.setToolTip(_translate("annotator_rhm", "press b"))
        self.lineButton.setText(_translate("annotator_rhm", "..."))
        self.tickButton.setToolTip(_translate("annotator_rhm", "press d"))
        self.tickButton.setText(_translate("annotator_rhm", "..."))
        self.moveButton.setToolTip(_translate("annotator_rhm", "press z"))
        self.moveButton.setText(_translate("annotator_rhm", "..."))
        self.redoButton.setToolTip(_translate("annotator_rhm", "press w"))
        self.redoButton.setText(_translate("annotator_rhm", "..."))
        self.zoomButton.setToolTip(_translate("annotator_rhm", "press a"))
        self.zoomButton.setText(_translate("annotator_rhm", "..."))
        self.deleteButton.setToolTip(_translate("annotator_rhm", "press x"))
        self.deleteButton.setText(_translate("annotator_rhm", "..."))
        self.commentButton.setToolTip(_translate("annotator_rhm", "press f"))
        self.commentButton.setText(_translate("annotator_rhm", "..."))
        self.undoButton.setToolTip(_translate("annotator_rhm", "press s"))
        self.undoButton.setText(_translate("annotator_rhm", "..."))
        self.crossButton.setToolTip(_translate("annotator_rhm", "press e"))
        self.crossButton.setText(_translate("annotator_rhm", "..."))
        self.boxButton.setToolTip(_translate("annotator_rhm", "press c"))
        self.boxButton.setText(_translate("annotator_rhm", "..."))
        self.modeLabel.setText(_translate("annotator_rhm", "move"))
        self.noAnswerButton.setText(_translate("annotator_rhm", "No answer"))
        self.rearrangePagesButton.setToolTip(_translate("annotator_rhm", "<html><head/><body><p>Find a missing page, discard irrelevant pages, rotate pages, etc</p></body></html>"))
        self.rearrangePagesButton.setText(_translate("annotator_rhm", "Adjust pages"))
        self.deltaButton.setText(_translate("annotator_rhm", "..."))



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

import hashlib
import json
import os
import uuid
import logging

log = logging.getLogger("servUI")

confdir = "serverConfiguration"


def validate(self, user, token):
    """Check the user's token is valid.

    Returns:
        bool
    """
    # log.debug("Validating user {}.".format(user))
    dbToken = self.DB.getUserToken(user)
    if not dbToken:
        log.warning('User "{}" tried a token but we have no such user!'.format(user))
        return False
    r = self.authority.validate_token(token, dbToken)
    # gives None/False/True
    if r is None:
        log.warning(
            'Malformed token from user "{}": client bug? malicious probing?'.format(
                user
            )
        )
    elif not r:
        log.info('User "{}" tried to use a stale or invalid token'.format(user))
    return bool(r)


def InfoShortName(self):
    if self.testSpec is None:
        return [False]
    else:
        return [True, self.testSpec["name"]]


def info_spec(self):
    """Return the exam specification.

    TODO: why not return None if no spec (yet)?

    Returns:
        tuple: first item flags success, second is the spec dict,
            see :func:`plom.specarser`.  Can fail when the server
            does not yet have a spec.  This function is not
            authenticated so strip the `privateSeed`.
    """
    if not self.testSpec:
        return False, None
    d = self.testSpec.copy()
    d.pop("privateSeed")
    return True, d


def reloadUsers(self, password):
    """Reload the user list."""
    # Check user is manager.
    if not self.authority.authoriseUser("manager", password):
        log.warning("Unauthorised attempt to reload users")
        return False
    log.info("Reloading the user list")
    # Load in the user list and check against existing user list for differences
    try:
        with open(os.path.join(confdir, "userList.json")) as data_file:
            newUserList = json.load(data_file)
            # for each user in the new list..
            for u in newUserList:
                if u not in self.userList:
                    # This is a new user - add them in.
                    self.userList[u] = newUserList[u]
                    self.authority.addUser(u, newUserList[u])
                    log.info("Adding new user = {}".format(u))
            # for each user in the old list..
            for u in self.userList:
                if u not in newUserList:
                    # this user has been removed
                    log.info("Removing user = {}".format(u))
                    # Anything out at user should go back on todo pile.
                    self.DB.resetUsersToDo(u)
                    # remove user's authorisation token.
                    self.authority.detoken(u)
    except FileNotFoundError:
        # TODO?  really not even return False?
        pass
    log.debug("Current user list = {}".format(list(self.userList.keys())))
    # return acknowledgement
    return True


def checkPassword(self, user, password):
    """Does user's password match the hashed one on file?"""
    hashed_pwd = self.DB.getUserPasswordHash(user)
    return self.authority.check_password(password, hashed_pwd)


def checkUserEnabled(self, user):
    return self.DB.isUserEnabled(user)


def giveUserToken(self, user, password, clientAPI, client_ver, remote_ip):
    """When a user requests authorisation
    They have sent their name and password
    first check if they are a valid user
    if so then anything that is recorded as out with that user
    should be reset as todo.
    Then pass them back the authorisation token
    (the password is only checked on first authorisation - since slow)
    """
    if clientAPI != self.API:
        return [
            False,
            "API"
            'Plom API mismatch: client "{}" =/= server "{}". Server version is "{}"; please check you have the right client.'.format(
                clientAPI, self.API, self.Version
            ),
        ]

    if not self.checkPassword(user, password):
        log.warning(
            'Invalid password login attempt by "{}" from {}, client ver "{}"'.format(
                user, remote_ip, client_ver
            )
        )
        return [False, "The name / password pair is not authorised"]

    if not self.checkUserEnabled(user):
        log.info('User "{}" logged in but account is disabled by manager'.format(user))
        return [
            False,
            "The name / password pair has been disabled. Contact your instructor.",
        ]

    # Now check if user already logged in - ie has token already.
    if self.DB.userHasToken(user):
        log.debug('User "{}" already has token'.format(user))
        return [False, "UHT", "User already has token."]
    # give user a token, and store the xor'd version.
    [clientToken, storageToken] = self.authority.create_token()
    self.DB.setUserToken(user, storageToken)
    # On token request also make sure anything "out" with that user is reset as todo.
    # We keep this here in case of client crash - todo's get reset on login and logout.
    self.DB.resetUsersToDo(user)
    log.info(
        'Authorising user "{}" from {}, client ver "{}"'.format(
            user, remote_ip, client_ver
        )
    )
    return [True, clientToken]


def setUserEnable(self, user, enableFlag):
    if enableFlag:
        self.DB.enableUser(user)
    else:
        self.DB.disableUser(user)
    return [True]


def createModifyUser(self, username, password):
    # basic sanity check of username / password
    if not self.authority.basic_user_password_check(username, password):
        return [False, "Username/Password fails basic checks."]
    if username == "HAL":  # Don't mess with HAL
        return [False, "I'm sorry, Dave. I'm afraid I can't do that."]
    # hash the password
    passwordHash = self.authority.create_password_hash(password)
    if self.DB.doesUserExist(username):  # user exists, so update password
        if self.DB.setUserPasswordHash(username, passwordHash):
            return [True, False]
        else:
            return [False, "Password update error."]
    else:  # user does not exist, so create them
        if self.DB.createUser(username, passwordHash):
            return [True, True]
        else:
            return [False, "User creation error."]


def closeUser(self, user):
    """Client is closing down their app, so remove the authorisation token"""
    log.info("Revoking auth token from user {}".format(user))
    self.DB.clearUserToken(user)
    # make sure all their out tasks are returned to "todo"
    self.DB.resetUsersToDo(user)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian

from .idReader import is_model_absent, calc_log_likelihood, download_or_train_model


def test_is_model_absent():
    assert is_model_absent() == True


def test_log_likelihood():
    import numpy as np

    num_digits = 8
    student_ids = [i for i in range(0, num_digits)]
    probabilities = [
        [0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0.5, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0.5, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0.5, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0.5, 0, 0],
    ]
    assert bool(
        np.isclose(
            calc_log_likelihood(student_ids, probabilities, num_digits),
            5.545177444479562,
            1e-7,
        )
    )


def test_download_or_train_model():
    assert is_model_absent() == True
    download_or_train_model()  # Although we cannot directly check the output of download_or_train_model, we can check that the is correct files are present and the function works as intended
    assert is_model_absent() == False



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

from datetime import datetime, timedelta

from peewee import *

from plom.rules import censorStudentNumber as censorID
from plom.rules import censorStudentName as censorName


import logging

log = logging.getLogger("DB")

from plom.db.tables import *

######################################################################


class PlomDB:
    def __init__(self, dbfile_name="plom.db"):
        # can't handle pathlib?
        plomdb.init(str(dbfile_name))

        with plomdb:
            plomdb.create_tables(
                [
                    User,
                    Image,
                    Bundle,
                    Test,
                    ##
                    Group,
                    IDGroup,
                    DNMGroup,
                    QGroup,
                    ##
                    TPage,
                    HWPage,
                    EXPage,
                    LPage,
                    UnknownPage,
                    CollidingPage,
                    DiscardedPage,
                    ##
                    AImage,
                    Annotation,
                    OldAnnotation,
                    ##
                    APage,
                    OAPage,
                    IDPage,
                    DNMPage,
                ]
            )
        log.info("Database initialised.")
        # check if HAL has been created
        if User.get_or_none(name="HAL") is None:
            User.create(
                name="HAL",
                password=None,
                last_activity=datetime.now(),
                last_action="Created",
            )
            log.info("User 'HAL' created to do all our automated tasks.")

    ########### User stuff #############
    from plom.db.db_user import (
        createUser,
        doesUserExist,
        setUserPasswordHash,
        getUserPasswordHash,
        isUserEnabled,
        enableUser,
        disableUser,
        setUserToken,
        clearUserToken,
        getUserToken,
        userHasToken,
        validateToken,
        getUserList,
        getUserDetails,
        resetUsersToDo,
    )

    from plom.db.db_create import (
        doesBundleExist,
        createNewBundle,
        createReplacementBundle,
        areAnyPapersProduced,
        nextqueue_position,
        createTest,
        addTPages,
        createIDGroup,
        createDNMGroup,
        createQGroup,
        getPageVersions,
        produceTest,
        id_paper,
    )

    from plom.db.db_upload import (
        createNewImage,
        attachImageToTPage,
        createNewHWPage,
        createNewLPage,
        uploadTestPage,
        uploadHWPage,
        uploadLPage,
        uploadUnknownPage,
        uploadCollidingPage,
        updateDNMGroup,
        updateIDGroup,
        cleanIDGroup,
        updateQGroup,
        cleanQGroup,
        updateGroupAfterUpload,
        checkTestScanned,
        updateTestAfterUpload,
        processUpdatedTests,
        getSIDFromTest,
        sidToTest,
        replaceMissingHWQuestion,
        replaceMissingTestPage,
        removeAllScannedPages,
    )

    from plom.db.db_manage import (
        getUnknownPageNames,
        getDiscardNames,
        getCollidingPageNames,
        getTPageImage,
        getHWPageImage,
        getEXPageImage,
        getLPageImage,
        getAllTestImages,
        getQuestionImages,
        getUnknownImage,
        testOwnersLoggedIn,
        moveUnknownToExtraPage,
        moveUnknownToHWPage,
        moveUnknownToTPage,
        checkTPage,
        removeUnknownImage,
        getDiscardImage,
        moveDiscardToUnknown,
        moveUnknownToCollision,
        getCollidingImage,
        removeCollidingImage,
        moveCollidingToTPage,
    )

    from plom.db.db_report import (
        RgetScannedTests,
        RgetIncompleteTests,
        RgetCompleteHW,
        RgetMissingHWQ,
        RgetUnusedTests,
        RgetIdentified,
        RgetProgress,
        RgetMarkHistogram,
        RgetMarked,
        RgetQuestionUserProgress,
        RgetCompletionStatus,
        RgetOutToDo,
        RgetStatus,
        RgetSpreadsheet,
        RgetOriginalFiles,
        RgetCoverPageInfo,
        RgetAnnotatedFiles,
        RgetMarkReview,
        RgetAnnotatedImage,
        RgetIDReview,
        RgetUserFullProgress,
    )

    from plom.db.db_identify import (
        IDcountAll,
        IDcountIdentified,
        IDgetNextTask,
        IDgiveTaskToClient,
        IDgetDoneTasks,
        IDgetImage,
        IDgetImageByNumber,
        IDdidNotFinish,
        ID_id_paper,
        IDgetImageFromATest,
        IDreviewID,
    )

    from plom.db.db_mark import (
        McountAll,
        McountMarked,
        MgetDoneTasks,
        MgetNextTask,
        MgiveTaskToClient,
        MdidNotFinish,
        MtakeTaskFromClient,
        MgetImages,
        MgetOneImageFilename,
        MgetOriginalImages,
        MsetTag,
        MgetWholePaper,
        MreviewQuestion,
        MrevertTask,
    )



#!/usr/bin/env python3

# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2019 Andrew Rechnitzer
# Copyright (C) 2020 Dryden Wiebe

"""
Produce a minimal archive of your test.

This is intended for archival purposes.  Note it would not be easy to
regrade a question within Plom using this archive.  If you ever
anticipate revisiting the grading of this test, you should backup the
entire directory structure.

Usage:
$ ./12_archive course year term

for example:
$ ./12_archive math253 2019 S1
"""

__copyright__ = "Copyright (C) 2019-2020 Colin B. Macdonald and others"
__credits__ = ["The Plom Project Developers"]
__license__ = "AGPL-3.0-or-later"

import os
import sys
import shutil

from plom import SpecParser
from plom.finish import CSVFilename


archivename = "{COURSE}_{YEAR}{TERM}_{SHORTNAME}"


if __name__ == "__main__":
    spec = SpecParser().spec
    basename = spec["name"]
    archivename = archivename.replace("{SHORTNAME}", basename)

    print("\n\nTODO: THIS SCRIPT NEEDS RETHINKING FOR 0.4!\n\n")

    # TODO: someday we can get this from spec file?
    # https://gitlab.com/plom/plom/issues/94
    if not len(sys.argv) == 4:
        print("ERROR: Incorrect command line...")
        print(__doc__)
        sys.exit(1)
    archivename = archivename.replace("{COURSE}", sys.argv[1])
    archivename = archivename.replace("{YEAR}", sys.argv[2])
    archivename = archivename.replace("{TERM}", sys.argv[3])

    print(
        """
This script tries to produce a minimal archive of your test:
    "{0}.zip"
This is intended for archival purposes.  Note it would not be easy to
regrade a question within Plom using this archive.  If you ever
anticipate revisiting the grading of this test, you should backup the
entire directory structure.
    """.format(
            archivename
        )
    )
    # input('Press Enter to continue...')

    all_ok = True

    print('Creating temporary directory "{0}"'.format(archivename))
    try:
        os.mkdir(archivename)
    except FileExistsError:
        print(
            'Directory "{0}" already exists: if you want to re-run this script, try deleting it first.'.format(
                archivename
            )
        )
        sys.exit(1)

    print("Archiving source pdf files")
    try:
        shutil.copytree(
            os.path.join("..", "build", "sourceVersions"),
            os.path.join(archivename, "sourceVersions"),
            symlinks=False,
        )
    except:
        print('  WARNING: could not archive "build/sourceVersions" directory')
        all_ok = False

    print("Archiving raw scans")
    try:
        shutil.copytree(
            os.path.join("..", "scanAndGroup", "scannedExams"),
            os.path.join(archivename, "scannedExams"),
            symlinks=False,
        )
    except:
        print('  WARNING: could not archive "scanAndGroup/scannedExams" directory')
        all_ok = False
    try:
        os.rmdir(os.path.join(archivename, "scannedExams", "png"))
    except:
        print(
            '  WARNING: could not remove supposedly-empty "scanningExams/png" directory'
        )
        all_ok = False

    t1 = os.path.isdir("reassembled")
    t2 = os.path.isdir("reassembled_ID_but_not_marked")
    if t1 and not t2:
        print('Archiving final graded pdf files ("reassembled")')
        shutil.copytree(
            "reassembled", os.path.join(archivename, "reassembled"), symlinks=False
        )
    elif t2 and not t1:
        print('Archiving final identified pdf files ("reassembled_ID_but_not_marked")')
        shutil.copytree(
            "reassembled_ID_but_not_marked",
            os.path.join(archivename, "reassembled_ID_but_not_marked"),
            symlinks=False,
        )
    elif t1 and t2:
        print(
            "  WARNING: found both reassembled and reassembled_ID_but_not_marked directories. Which should be archived?"
        )
        all_ok = False
    else:
        print("  WARNING: cannot find any reassembled papers.")
        all_ok = False

    print("Archiving metadata and miscellanea")
    shutil.copy2(CSVFilename, archivename)
    shutil.copy2(os.path.join("..", "resources", "testSpec.json"), archivename)
    # which version was used for each exam number
    shutil.copy2(os.path.join("..", "resources", "examsProduced.json"), archivename)
    # the mapping b/w scan file and paper number
    shutil.copy2(os.path.join("..", "resources", "examsScanned.json"), archivename)
    # the mapping b/w paper number and student
    shutil.copy2(os.path.join("..", "resources", "examsIdentified.json"), archivename)
    # who did the marking
    shutil.copy2(os.path.join("..", "resources", "groupImagesMarked.json"), archivename)
    # TODO: information about grading times?

    with open(os.path.join(archivename, "README.txt"), "w") as file:
        file.write(
            """Plom Archive File
=================

Explanations:

  * reassembled: these are the final graded pdf files, by student number.

  * reassembled_ID_but_not_marked: these are the final identified (but not marked) pdf files, by student number. This probably means the marking was done on-paper before the papers were scanned. Note that only one of "reassembled" and "reassembled_ID_but_not_marked" should be present.

  * sourceVersions: the original blank pdfs for the test/exam.  Each
    student's test is some combination of these.

  * scannedExams: the raw unmarked files directly from the scanner.

  * The various .json files can be used to find students' papers in the raw
    scans.
"""
        )

    print("Creating zip file")
    fn = shutil.make_archive(archivename, "zip", base_dir=archivename)
    print('  created "{0}"'.format(fn))

    print("Removing temp directory")
    shutil.rmtree(archivename)

    print("\nFinished!  The zip file is:\n  {0}".format(fn))
    print("  (you may want to rename this)")
    if not all_ok:
        print("... but there were some warnings, see above")



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Vala Vakilian

from aiohttp import web, MultipartWriter, MultipartReader

from .routeutils import authenticate_by_token, authenticate_by_token_required_fields
from .routeutils import validate_required_fields, log_request
from .routeutils import log


class MarkHandler:
    def __init__(self, plomServer):
        self.server = plomServer

    # @routes.get("/MK/maxMark")
    @authenticate_by_token_required_fields(["q", "v"])
    def MgetQuestionMark(self, data, request):
        """Retreive the maximum mark for a question.

        Respond with status 200/416.

        Args:
            data (dict): Dictionary including user data in addition to question number
                and test version.
            request (aiohttp.web_response.Response): GET /MK/maxMark request object.

        Returns:
            aiohttp.web_response.Response: Response object which has the
                maximum mark for the question.
        """
        question_number = data["q"]
        test_version = data["v"]
        valid_question, maximum_mark_response = self.server.MgetQuestionMax(
            question_number, test_version
        )

        if valid_question:
            return web.json_response(maximum_mark_response, status=200)
        elif maximum_mark_response == "QE":  # Check if the question was out of range
            # pg out of range
            return web.Response(
                text="Question out of range - please check before trying again.",
                status=416,
            )
        elif maximum_mark_response == "VE":  # Check if the version was out of range
            # version our of range
            return web.Response(
                text="Version out of range - please check before trying again.",
                status=416,
            )

    # @routes.get("/MK/progress")
    @authenticate_by_token_required_fields(["q", "v"])
    def MprogressCount(self, data, request):
        """Respond with the number of marked questions and the total questions tasks for user.

        Respond with status 200.

        Args:
            data (dict): Dictionary including user data in addition to question number
                and test version.
            request (aiohttp.web_request.Request): Request of type GET /MK/progress.

        Returns:
            aiohttp.web_response.Response: Includes the number of marked
                tasks and the total number of marked/unmarked tasks.
        """
        return web.json_response(
            self.server.MprogressCount(data["q"], data["v"]), status=200
        )

    # @routes.get("/MK/tasks/complete")
    @authenticate_by_token_required_fields(["user", "q", "v"])
    def MgetDoneTasks(self, data, request):
        """Retrieve data for questions which have already been graded by the user.

        Respond with status 200.

        Args:
            data (dict): Dictionary including user data in addition to
                question number and test version.
            request (aiohttp.web_response.Response): GET /MK/tasks/complete request object.

        Returns:
            aiohttp.web_response.Response: A response object including a
                list of lists with the already processed questions. The
                list involves the question string, question mark, time
                spent grading and tag string.
        """
        # return the completed list
        return web.json_response(
            self.server.MgetDoneTasks(data["user"], data["q"], data["v"]),
            status=200,
        )

    # @routes.get("/MK/tasks/available")
    @authenticate_by_token_required_fields(["q", "v"])
    def MgetNextTask(self, data, request):
        """Respond with the next task/question's string.

        Respond with status 200/204.

        Args:
            data (dict): Dictionary including user data in addition to
                question number and test version.
            request (aiohttp.web_request.Request): Request of type GET /MK/tasks/available.

        Returns:
            aiohttp.web_response.Response: A response which includes the next question's string.
                For example: q0013g1
        """
        next_task_response = self.server.MgetNextTask(data["q"], data["v"])

        next_task_available = next_task_response[0]

        # returns [True, task] or [False]
        if next_task_available:
            next_task_code = next_task_response[1]
            return web.json_response(next_task_code, status=200)
        else:
            return web.Response(status=204)  # no papers left

    # @routes.get("/MK/latex")
    @authenticate_by_token_required_fields(["user", "fragment"])
    def MlatexFragment(self, data, request):
        """Return the latex image for the string included in the request.

        Respond with status 200/406.

        Args:
            data (dict): Includes the user/token and latex string fragment.
            request (aiohttp.web_request.Request): Request of type GET /MK/latex.

        Returns:
            aiohttp.web_fileresponse.FileResponse: A response which includes the image for
                the latex string.
        """
        latex_response = self.server.MlatexFragment(data["user"], data["fragment"])
        latex_valid = latex_response[0]

        if latex_valid:
            latex_image_path = latex_response[1]
            return web.FileResponse(latex_image_path, status=200)
        else:
            return web.Response(status=406)  # a latex error

    # @routes.patch("/MK/tasks/{task}")
    @authenticate_by_token_required_fields(["user"])
    def MclaimThisTask(self, data, request):
        """Take task number in request and return the task/question's image data as a response.

        Respond with status 200/204.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): PATCH /MK/tasks/`question code` request object.
                This request object will include the task code.

        Returns:
            aiohttp.web_response.json_response: metadata about the images.
        """

        task_code = request.match_info["task"]
        # returns either
        #   [True, image_metadata, tags, integrity_check]
        #   [False]
        retvals = self.server.MclaimThisTask(data["user"], task_code)

        if not retvals[0]:
            return web.Response(status=204)  # that task already taken.
        return web.json_response(retvals[1:], status=200)

    # @routes.delete("/MK/tasks/{task}")
    @authenticate_by_token_required_fields(["user"])
    def MdidNotFinishTask(self, data, request):
        """Assign tasks that are not graded (untouched) as unfinished in the database.

        Respond with status 200.

        Args:
            data (dict): Includes the user/token and task code.
            request (aiohttp.web_request.Request): Request of type DELETE /MK/tasks/`question code`.

        Returns:
            aiohttp.web_response.Response: Returns a success status indicating the task is unfinished.
        """

        task_code = request.match_info["task"]
        self.server.MdidNotFinish(data["user"], task_code)

        return web.json_response(status=200)

    # @routes.put("/MK/tasks/{task}")
    async def MreturnMarkedTask(self, request):
        """Save the graded/processes task, extract data and save to database.

        This function also responds with the number of done tasks and the total number of tasks.
        The returned statement is similar to MprogressCount.
        Respond with status 200/400/401/406/409/410.
        Log activity.

        Args:
            request (aiohttp.web_request.Request): Request of type
                PUT /MK/tasks/`question code` which includes a multipart
                object indication the marked test data. This request will
                include 3 parts including [metadata, image, plom-file].

        Returns:
            aiohttp.web_response.Response: Responses with a list including the number of
                graded tasks and the overall number of tasks.
        """

        log_request("MreturnMarkedTask", request)

        reader = MultipartReader.from_response(request)

        # Dealing with the metadata.
        task_metadata_object = await reader.next()

        if task_metadata_object is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        task_metadata = await task_metadata_object.json()

        # Validate that the dictionary has these fields.
        if not validate_required_fields(
            task_metadata,
            [
                "user",
                "token",
                "comments",
                "pg",
                "ver",
                "score",
                "mtime",
                "tags",
                "md5sum",
                "integrity_check",
                "image_md5s",
            ],
        ):
            return web.Response(status=400)
        # Validate username and token.
        if not self.server.validate(task_metadata["user"], task_metadata["token"]):
            return web.Response(status=401)

        comments = task_metadata["comments"]  # List of comments.
        task_code = request.match_info["task"]  # Task code.

        # Note: if user isn't validated, we don't parse their binary junk
        # TODO: is it safe to abort during a multi-part thing?

        # Dealing with the image.
        task_image_object = await reader.next()

        if task_image_object is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        task_image = await task_image_object.read()

        # Dealing with the plom_file.
        plom_file_object = await reader.next()

        if plom_file_object is None:  # weird error
            return web.Response(status=406)  # should have sent 3 parts
        plomdat = await plom_file_object.read()

        marked_task_status = self.server.MreturnMarkedTask(
            task_metadata["user"],
            task_code,
            int(task_metadata["pg"]),
            int(task_metadata["ver"]),
            int(task_metadata["score"]),
            task_image,
            plomdat,
            comments,
            int(task_metadata["mtime"]),
            task_metadata["tags"],
            task_metadata["md5sum"],
            task_metadata["integrity_check"],
            task_metadata["image_md5s"],
        )
        # marked_task_status = either [True, Num Done tasks, Num Totalled tasks] or [False] if error.

        if marked_task_status[0]:
            num_done_tasks = marked_task_status[1]
            total_num_tasks = marked_task_status[2]
            return web.json_response([num_done_tasks, total_num_tasks], status=200)
        else:
            if marked_task_status[1] == "no_such_task":
                log.warning("Returning with error 410 = {}".format(marked_task_status))
                return web.Response(status=410)
            elif marked_task_status[1] == "not_owner":
                log.warning("Returning with error 409 = {}".format(marked_task_status))
                return web.Response(status=409)
            elif marked_task_status[1] == "integrity_fail":
                log.warning("Returning with error 406 = {}".format(marked_task_status))
                return web.Response(status=406)
            else:
                log.warning("Returning with error 400 = {}".format(marked_task_status))
                return web.Response(status=400)

    # @routes.get("/MK/images/{task}")
    @authenticate_by_token_required_fields(["user", "integrity_check"])
    def MgetImages(self, data, request):
        """Return underlying image data and annotations of a question/task.

        Main API call for the client to get the image data (original and annotated).
        Respond with status 200/409.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /MK/images/"task code"
                which the task code is extracted from.

        Returns:
            aiohttp.web_response.Response: A response which includes the multipart writer object
                wrapping the task images.
        """

        task_code = request.match_info["task"]
        results = self.server.MgetImages(
            data["user"], task_code, data["integrity_check"]
        )
        # Format is one of:
        # [False, error]
        # [True, image_data]
        # [True, image_data, annotated_fname, plom_filename]
        if not results[0]:
            if results[1] == "owner":
                return web.Response(status=409)  # someone else has that task_image
            elif results[1] == "integrity_fail":
                return web.Response(status=406)  # task changed
            elif results[1] == "no_such_task":
                return web.Response(status=410)  # task deleted
            else:
                return web.Response(status=400)  # some other error

        with MultipartWriter("imageAnImageAndPlom") as multipart_writer:
            image_metadata = results[1]
            files = []
            # append the annotated_fname, plom_filename if present
            files.extend(results[2:])

            multipart_writer.append_json(image_metadata)
            for file_name in files:
                multipart_writer.append(open(file_name, "rb"))
        return web.Response(body=multipart_writer, status=200)

    # @routes.get(...)
    @authenticate_by_token_required_fields(["user"])
    def MgetOneImage(self, data, request):
        """Return one image from the database.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request)

        Returns:
            aiohttp.web_response.Response: the binary image data, or
                a 404 response if no such image, or a 409 if wrong
                md5sum saniity check was provided.
        """
        task_code = request.match_info["task"]
        image_id = request.match_info["image_id"]
        md5sum = request.match_info["md5sum"]

        r = self.server.DB.MgetOneImageFilename(
            data["user"], task_code, image_id, md5sum
        )
        if not r[0]:
            if r[1] == "no such image":
                return web.Response(status=404)
            elif r[1] == "wrong md5sum":
                return web.Response(status=409)
            else:
                return web.Response(status=500)
        filename = r[1]
        return web.FileResponse(filename, status=200)

    # @routes.get("/MK/originalImage/{task}")
    @authenticate_by_token_required_fields([])
    def MgetOriginalImages(self, data, request):
        """Return the non-graded original images for a task/question.

        Respond with status 200/204.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type
                GET /MK/originalImages/`task code` which the task code
                is extracted from.

        Returns:
            aiohttp.web_response.Response: A response object with includes the multipart objects
                which wrap this task/question's original (ungraded) images.
        """

        task = request.match_info["task"]
        get_image_results = self.server.MgetOriginalImages(task)
        # returns either [True, md5s, fname1, fname2,... ] or [False]
        image_return_success = get_image_results[0]

        if not image_return_success:
            return web.Response(status=204)  # no content there

        original_image_paths = get_image_results[1:]
        with MultipartWriter("images") as multipart_writer:
            for file_nam in original_image_paths:
                multipart_writer.append(open(file_nam, "rb"))
        return web.Response(body=multipart_writer, status=200)

    # @routes.patch("/MK/tags/{task}")
    @authenticate_by_token_required_fields(["user", "tags"])
    def MsetTag(self, data, request):
        """Set tag for a task.

        Respond with status 200/409.

        Args:
            data (dict): A dictionary having the user/token in addition to the tag string.
                Request object also incudes the task code.
            request (aiohttp.web_request.Request): PATCH /MK/tags/`task_code` type request.

        Returns:
            aiohttp.web_response.Response: Empty status response indication is adding
                the tag was successful.
        """

        task_code = request.match_info["task"]
        set_tag_success = self.server.MsetTag(data["user"], task_code, data["tags"])

        if set_tag_success:
            return web.Response(status=200)
        else:
            return web.Response(status=409)  # Task does not belong to this user.

    # @routes.get("/MK/whole/{number}")
    @authenticate_by_token_required_fields([])
    def MgetWholePaper(self, data, request):
        """Return the entire paper which includes the given question.

        Respond with status 200/404.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): GET /MK/whole/`test_number`/`question_number`.

        Returns:
            aiohttp.web_response.Response: Responds with a multipart
                writer which includes all the images for the exam which
                includes this question.
        """
        test_number = request.match_info["number"]
        question_number = request.match_info["question"]

        # return [True, pageData, f1, f2, f3, ...] or [False]
        # 1. True/False for operation status.
        # 2. A list of lists, documented elsewhere (TODO: I hope)
        # 3. 3rd element onward: paths for each page of the paper in server.
        whole_paper_response = self.server.MgetWholePaper(test_number, question_number)

        if not whole_paper_response[0]:
            return web.Response(status=404)

        with MultipartWriter("images") as multipart_writer:
            pages_data = whole_paper_response[1]
            all_pages_paths = whole_paper_response[2:]
            multipart_writer.append_json(pages_data)
            for file_name in all_pages_paths:
                multipart_writer.append(open(file_name, "rb"))
        return web.Response(body=multipart_writer, status=200)

    # @routes.get("/MK/TMP/whole/{number}/{question}")
    @authenticate_by_token_required_fields([])
    def MgetWholePaperMetadata(self, data, request):
        """Return the metadata for all images associated with a paper

        Respond with status 200/404.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): GET /MK/whole/`test_number`/`question_number`.

        Returns:
            aiohttp.web_response.Response: JSON data, a list of lists
                where each list in the form documented below.

        Each row of the data looks like:
           `[name, md5, true/false, pos_in_current_annotation, image_id]`
        """
        test_number = request.match_info["number"]
        # TODO: who cares about this?
        question_number = request.match_info["question"]

        # return [True, pageData, f1, f2, f3, ...] or [False]
        # 1. True/False for operation status.
        # 2. A list of lists, documented elsewhere (TODO: I hope)
        # 3. 3rd element onward: paths for each page of the paper in server.
        r = self.server.MgetWholePaper(test_number, question_number)

        if not r[0]:
            return web.Response(status=404)

        pages_data = r[1]
        # We just discard this, its legacy from previous API call
        # TODO: fold into the pages_data; then we can sanity check it when it comes back!
        all_pages_paths = r[2:]
        return web.json_response(pages_data, status=200)

    # @routes.get("/MK/allMax")
    @authenticate_by_token
    def MgetAllMax(self):
        """Respond with information on max mark possible for each question in the exam.

        Respond with status 200/404.

        Returns:
            aiohttp.web_response.Response: A response which includes a dictionary
                for the highest mark possible for each question of the exam.
        """
        return web.json_response(self.server.MgetAllMax(), status=200)

    # @routes.patch("/MK/review")
    @authenticate_by_token_required_fields(["testNumber", "questionNumber", "version"])
    def MreviewQuestion(self, data, request):
        """Confirm the question review done on plom-manager.

        Respond with status 200/404.

        Args:
            data (dict): Dictionary including user data in addition to question number
                and test version.
            request (aiohttp.web_request.Request): Request of type PATCH /MK/review .

        Returns:
            aiohttp.web_response.Response: Empty status response indication if the question
                review was successful.
        """

        if self.server.MreviewQuestion(
            data["testNumber"], data["questionNumber"], data["version"]
        ):
            return web.Response(status=200)
        else:
            return web.Response(status=404)

    # @routes.patch("/MK/revert/{task}")
    # TODO: Deprecated.
    # TODO: Should be removed.
    @authenticate_by_token_required_fields(["user"])
    def MrevertTask(self, data, request):
        # only manager can do this
        task = request.match_info["task"]
        if not data["user"] == "manager":
            return web.Response(status=401)  # malformed request.
        rval = self.server.MrevertTask(task)
        if rval[0]:
            return web.Response(status=200)
        elif rval[1] == "NAC":  # nothing to be done here.
            return web.Response(status=204)
        else:  # cannot find that task
            return web.Response(status=404)

    def setUpRoutes(self, router):
        """Adds the response functions to the router object.

        Args:
            router (aiohttp.web_urldispatcher.UrlDispatcher): Router object which we will add the response functions to.
        """
        router.add_get("/MK/allMax", self.MgetAllMax)
        router.add_get("/MK/maxMark", self.MgetQuestionMark)
        router.add_get("/MK/progress", self.MprogressCount)
        router.add_get("/MK/tasks/complete", self.MgetDoneTasks)
        router.add_get("/MK/tasks/available", self.MgetNextTask)
        router.add_get("/MK/latex", self.MlatexFragment)
        router.add_patch("/MK/tasks/{task}", self.MclaimThisTask)
        router.add_delete("/MK/tasks/{task}", self.MdidNotFinishTask)
        router.add_put("/MK/tasks/{task}", self.MreturnMarkedTask)
        router.add_get("/MK/images/{task}", self.MgetImages)
        router.add_get("/MK/images/{task}/{image_id}/{md5sum}", self.MgetOneImage)
        router.add_get("/MK/originalImages/{task}", self.MgetOriginalImages)
        router.add_patch("/MK/tags/{task}", self.MsetTag)
        router.add_get("/MK/whole/{number}/{question}", self.MgetWholePaper)
        router.add_get("/MK/TMP/whole/{number}/{question}", self.MgetWholePaperMetadata)
        router.add_patch("/MK/review", self.MreviewQuestion)
        router.add_patch("/MK/revert/{task}", self.MrevertTask)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian

# https://www.digitalocean.com/community/tutorials/how-to-build-a-neural-network-to-recognize-handwritten-digits-with-tensorflow
# and https://www.tensorflow.org/api_docs/python/tf/keras/datasets/mnist/load_data
# https://www.tensorflow.org/tutorials/estimators/cnn
# and then hacked to work with tensorflow 2.

# tensorflow bits and pieces
import tensorflow as tf
from tensorflow import keras


def train_model():
    """Grab the mnist data to train against."""

    print("Getting mnist handwritten digit data set.")
    mnist = tf.keras.datasets.mnist
    (X_train, y_train), (X_test, y_test) = mnist.load_data()

    # define the model using keras Sequential
    # https://keras.io/guides/sequential_model/
    print("Build and compile digit-recognition model")
    model = keras.Sequential(
        [
            keras.layers.Flatten(input_shape=(28, 28)),
            keras.layers.Reshape((28, 28, 1)),
            keras.layers.Conv2D(
                filters=32, kernel_size=[5, 5], activation=tf.nn.relu, padding="same"
            ),
            keras.layers.MaxPooling2D(pool_size=[2, 2], strides=2),
            keras.layers.Conv2D(
                filters=64, kernel_size=[5, 5], activation=tf.nn.relu, padding="same"
            ),
            keras.layers.MaxPooling2D(pool_size=[2, 2], strides=2),
            keras.layers.Flatten(),
            keras.layers.Dense(units=1024, activation=tf.nn.relu),
            keras.layers.Dropout(rate=0.4),
            keras.layers.Dense(units=10, activation=tf.nn.softmax),
        ]
    )

    # Compile it
    model.compile(
        optimizer="adam",
        loss="sparse_categorical_crossentropy",
        metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
    )

    print("Start training model")
    # Train it
    model.fit(X_train, y_train, epochs=8)

    test_loss, test_acc = model.evaluate(X_test, y_test)
    print("Accuracy of model on mnist test-set:", test_acc)

    print("Save the model")
    model.save("model_cache")


if __name__ == "__main__":
    train_model()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './qtCreatorFiles/ui_chooser.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Chooser(object):
    def setupUi(self, Chooser):
        Chooser.setObjectName("Chooser")
        Chooser.resize(540, 572)
        self.verticalLayout = QtWidgets.QVBoxLayout(Chooser)
        self.verticalLayout.setObjectName("verticalLayout")
        self.serverGBox = QtWidgets.QGroupBox(Chooser)
        self.serverGBox.setEnabled(True)
        self.serverGBox.setObjectName("serverGBox")
        self.gridLayout = QtWidgets.QGridLayout(self.serverGBox)
        self.gridLayout.setObjectName("gridLayout")
        self.infoLabel = QtWidgets.QLabel(self.serverGBox)
        self.infoLabel.setText("")
        self.infoLabel.setObjectName("infoLabel")
        self.gridLayout.addWidget(self.infoLabel, 3, 1, 1, 3)
        self.infoLabelStatic = QtWidgets.QLabel(self.serverGBox)
        self.infoLabelStatic.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.infoLabelStatic.setObjectName("infoLabelStatic")
        self.gridLayout.addWidget(self.infoLabelStatic, 3, 0, 1, 1)
        self.serverLabel = QtWidgets.QLabel(self.serverGBox)
        self.serverLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.serverLabel.setObjectName("serverLabel")
        self.gridLayout.addWidget(self.serverLabel, 1, 0, 1, 1)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.getServerInfoButton = QtWidgets.QPushButton(self.serverGBox)
        self.getServerInfoButton.setObjectName("getServerInfoButton")
        self.horizontalLayout_5.addWidget(self.getServerInfoButton)
        self.gridLayout.addLayout(self.horizontalLayout_5, 2, 2, 1, 2)
        self.mportLabel = QtWidgets.QLabel(self.serverGBox)
        self.mportLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.mportLabel.setObjectName("mportLabel")
        self.gridLayout.addWidget(self.mportLabel, 2, 0, 1, 1)
        self.serverLE = QtWidgets.QLineEdit(self.serverGBox)
        self.serverLE.setObjectName("serverLE")
        self.gridLayout.addWidget(self.serverLE, 1, 1, 1, 3)
        self.mportSB = QtWidgets.QSpinBox(self.serverGBox)
        self.mportSB.setMaximum(65535)
        self.mportSB.setProperty("value", 41984)
        self.mportSB.setObjectName("mportSB")
        self.gridLayout.addWidget(self.mportSB, 2, 1, 1, 1)
        self.gridLayout.setColumnStretch(0, 1)
        self.gridLayout.setColumnStretch(1, 1)
        self.gridLayout.setColumnStretch(2, 1)
        self.gridLayout.setColumnStretch(3, 1)
        self.verticalLayout.addWidget(self.serverGBox)
        self.markGBox = QtWidgets.QGroupBox(Chooser)
        self.markGBox.setEnabled(True)
        self.markGBox.setObjectName("markGBox")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.markGBox)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.vlabel = QtWidgets.QLabel(self.markGBox)
        self.vlabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.vlabel.setObjectName("vlabel")
        self.gridLayout_2.addWidget(self.vlabel, 0, 2, 1, 1)
        self.pgLayout = QtWidgets.QHBoxLayout()
        self.pgLayout.setObjectName("pgLayout")
        self.pgSB = QtWidgets.QSpinBox(self.markGBox)
        self.pgSB.setProperty("value", 1)
        self.pgSB.setObjectName("pgSB")
        self.pgLayout.addWidget(self.pgSB)
        self.pgDrop = QtWidgets.QComboBox(self.markGBox)
        self.pgDrop.setObjectName("pgDrop")
        self.pgLayout.addWidget(self.pgDrop)
        self.gridLayout_2.addLayout(self.pgLayout, 0, 1, 1, 1)
        self.pgLabel = QtWidgets.QLabel(self.markGBox)
        self.pgLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.pgLabel.setObjectName("pgLabel")
        self.gridLayout_2.addWidget(self.pgLabel, 0, 0, 1, 1)
        self.vLayout = QtWidgets.QHBoxLayout()
        self.vLayout.setObjectName("vLayout")
        self.vSB = QtWidgets.QSpinBox(self.markGBox)
        self.vSB.setProperty("value", 1)
        self.vSB.setObjectName("vSB")
        self.vLayout.addWidget(self.vSB)
        self.vDrop = QtWidgets.QComboBox(self.markGBox)
        self.vDrop.setObjectName("vDrop")
        self.vLayout.addWidget(self.vDrop)
        self.gridLayout_2.addLayout(self.vLayout, 0, 3, 1, 1)
        self.gridLayout_2.setColumnStretch(0, 1)
        self.gridLayout_2.setColumnStretch(1, 1)
        self.gridLayout_2.setColumnStretch(2, 1)
        self.gridLayout_2.setColumnStretch(3, 1)
        self.verticalLayout.addWidget(self.markGBox)
        self.userGBox = QtWidgets.QGroupBox(Chooser)
        self.userGBox.setEnabled(True)
        self.userGBox.setObjectName("userGBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.userGBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label = QtWidgets.QLabel(self.userGBox)
        self.label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.gridLayout_3.addWidget(self.label, 0, 0, 1, 1)
        self.userLE = QtWidgets.QLineEdit(self.userGBox)
        self.userLE.setObjectName("userLE")
        self.gridLayout_3.addWidget(self.userLE, 0, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.userGBox)
        self.label_2.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_2.setObjectName("label_2")
        self.gridLayout_3.addWidget(self.label_2, 1, 0, 1, 1)
        self.passwordLE = QtWidgets.QLineEdit(self.userGBox)
        self.passwordLE.setEchoMode(QtWidgets.QLineEdit.Password)
        self.passwordLE.setObjectName("passwordLE")
        self.gridLayout_3.addWidget(self.passwordLE, 1, 1, 1, 1)
        self.gridLayout_3.setColumnStretch(0, 1)
        self.gridLayout_3.setColumnStretch(1, 3)
        self.verticalLayout.addWidget(self.userGBox)
        self.settingsBox = QtWidgets.QGroupBox(Chooser)
        self.settingsBox.setObjectName("settingsBox")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.settingsBox)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.fontLabel = QtWidgets.QLabel(self.settingsBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fontLabel.sizePolicy().hasHeightForWidth())
        self.fontLabel.setSizePolicy(sizePolicy)
        self.fontLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.fontLabel.setObjectName("fontLabel")
        self.gridLayout_4.addWidget(self.fontLabel, 0, 0, 1, 1)
        self.fontSB = QtWidgets.QSpinBox(self.settingsBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fontSB.sizePolicy().hasHeightForWidth())
        self.fontSB.setSizePolicy(sizePolicy)
        self.fontSB.setMinimum(2)
        self.fontSB.setMaximum(24)
        self.fontSB.setProperty("value", 10)
        self.fontSB.setObjectName("fontSB")
        self.gridLayout_4.addWidget(self.fontSB, 0, 1, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.fontButton = QtWidgets.QPushButton(self.settingsBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fontButton.sizePolicy().hasHeightForWidth())
        self.fontButton.setSizePolicy(sizePolicy)
        self.fontButton.setObjectName("fontButton")
        self.horizontalLayout_3.addWidget(self.fontButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.optionsButton = QtWidgets.QPushButton(self.settingsBox)
        self.optionsButton.setObjectName("optionsButton")
        self.horizontalLayout_3.addWidget(self.optionsButton)
        self.gridLayout_4.addLayout(self.horizontalLayout_3, 0, 2, 1, 2)
        self.gridLayout_4.setColumnStretch(0, 1)
        self.gridLayout_4.setColumnStretch(1, 1)
        self.gridLayout_4.setColumnStretch(2, 1)
        self.gridLayout_4.setColumnStretch(3, 1)
        self.verticalLayout.addWidget(self.settingsBox)
        self.taskGBox = QtWidgets.QGroupBox(Chooser)
        self.taskGBox.setObjectName("taskGBox")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.taskGBox)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.markButton = QtWidgets.QPushButton(self.taskGBox)
        self.markButton.setObjectName("markButton")
        self.horizontalLayout.addWidget(self.markButton)
        self.identifyButton = QtWidgets.QPushButton(self.taskGBox)
        self.identifyButton.setObjectName("identifyButton")
        self.horizontalLayout.addWidget(self.identifyButton)
        self.totalButton = QtWidgets.QPushButton(self.taskGBox)
        self.totalButton.setEnabled(False)
        self.totalButton.setObjectName("totalButton")
        self.horizontalLayout.addWidget(self.totalButton)
        self.closeButton = QtWidgets.QPushButton(self.taskGBox)
        self.closeButton.setObjectName("closeButton")
        self.horizontalLayout.addWidget(self.closeButton)
        self.verticalLayout.addWidget(self.taskGBox)

        self.retranslateUi(Chooser)
        QtCore.QMetaObject.connectSlotsByName(Chooser)
        Chooser.setTabOrder(self.serverLE, self.mportSB)
        Chooser.setTabOrder(self.mportSB, self.getServerInfoButton)
        Chooser.setTabOrder(self.getServerInfoButton, self.pgSB)
        Chooser.setTabOrder(self.pgSB, self.pgDrop)
        Chooser.setTabOrder(self.pgDrop, self.vSB)
        Chooser.setTabOrder(self.vSB, self.vDrop)
        Chooser.setTabOrder(self.vDrop, self.userLE)
        Chooser.setTabOrder(self.userLE, self.passwordLE)
        Chooser.setTabOrder(self.passwordLE, self.fontSB)
        Chooser.setTabOrder(self.fontSB, self.fontButton)
        Chooser.setTabOrder(self.fontButton, self.optionsButton)
        Chooser.setTabOrder(self.optionsButton, self.markButton)
        Chooser.setTabOrder(self.markButton, self.identifyButton)
        Chooser.setTabOrder(self.identifyButton, self.totalButton)
        Chooser.setTabOrder(self.totalButton, self.closeButton)

    def retranslateUi(self, Chooser):
        _translate = QtCore.QCoreApplication.translate
        Chooser.setWindowTitle(_translate("Chooser", "Plom Client"))
        self.serverGBox.setTitle(_translate("Chooser", "Server Information"))
        self.infoLabelStatic.setText(_translate("Chooser", "Info:"))
        self.serverLabel.setText(_translate("Chooser", "Server name:"))
        self.getServerInfoButton.setText(_translate("Chooser", "&Validate server"))
        self.mportLabel.setText(_translate("Chooser", "Port:"))
        self.serverLE.setText(_translate("Chooser", "127.0.0.1"))
        self.markGBox.setTitle(_translate("Chooser", "Marking information"))
        self.vlabel.setText(_translate("Chooser", "Version:"))
        self.pgLabel.setText(_translate("Chooser", "Question"))
        self.userGBox.setTitle(_translate("Chooser", "User Information"))
        self.label.setText(_translate("Chooser", "Username:"))
        self.label_2.setText(_translate("Chooser", "Password:"))
        self.settingsBox.setTitle(_translate("Chooser", "Settings"))
        self.fontLabel.setText(_translate("Chooser", "Font size:"))
        self.fontButton.setText(_translate("Chooser", "Set font"))
        self.optionsButton.setText(_translate("Chooser", "More &options..."))
        self.taskGBox.setTitle(_translate("Chooser", "Choose task"))
        self.markButton.setText(_translate("Chooser", "&Mark"))
        self.identifyButton.setText(_translate("Chooser", "&Identify"))
        self.totalButton.setText(_translate("Chooser", "Future use"))
        self.closeButton.setText(_translate("Chooser", "&Close"))



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald


"""Exceptions for the Plom software.

Serious exceptions are for unexpected things that we probably cannot
sanely or safely recover from.  Benign are for signaling expected (or
at least not unexpected) situations.
"""


class PlomException(Exception):
    """Catch-all parent of all Plom-related exceptions."""

    pass


class PlomSeriousException(PlomException):
    """Serious or unexpected problems that are generally not recoverable."""

    pass


class PlomBenignException(PlomException):
    """A not-unexpected situation, often signaling an error condition."""

    pass


class PlomAPIException(PlomBenignException):
    pass


class PlomConflict(PlomBenignException):
    """The action was contradictory to info already in the system."""

    pass


class PlomNoMoreException(PlomBenignException):
    pass


class PlomRangeException(PlomBenignException):
    pass


class PlomAuthenticationException(PlomBenignException):
    """You are not authenticated, with precisely that as the default message."""

    def __init__(self, msg=None):
        if not msg:
            msg = "You are not authenticated."
        super().__init__(msg)


class PlomTakenException(PlomBenignException):
    pass


class PlomLatexException(PlomBenignException):
    pass


class PlomExistingLoginException(PlomBenignException):
    pass


class PlomOwnersLoggedInException(PlomBenignException):
    pass


class PlomTaskChangedError(PlomBenignException):
    pass


class PlomTaskDeletedError(PlomBenignException):
    pass



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

"""
Utilities for canned users and auto-generated (poor) passwords
"""

import secrets
from random import sample

words = """
about above across act active activity add afraid after again age ago agree
air all alone along already always am amount an and angry another answer any
anyone anything anytime appear apple are area arm army around arrive art as ask
at attack aunt autumn away baby back bad bag ball bank base basket bath be bean
bear bed bedroom beer behave before begin behind bell below besides best better
between big bird birth birthday bit bite black bleed block blood blow blue board
boat body boil bone book border born borrow both bottle bottom bowl box boy
branch brave bread break breathe bridge bright bring brother brown brush build
burn business bus busy but buy by cake call can candle cap car card care careful
careless carry case cat catch central century certain chair chance change chase
cheap cheese chicken child children choice choose circle city class clever clean
clear climb clock cloth clothes cloud cloudy close coffee coat coin cold collect
colour comb common compare come complete computer continue control cook cool
copper corn corner correct cost contain count country course cover crash cross
cry cup cupboard cut dance dark daughter day dead decide decrease deep deer
depend desk destroy develop die dinner dirty discover dish do dog door double
down draw dream dress drink drive drop dry duck dust duty each ear early earn
earth east easy eat effect egg eight either electric elephant else empty end
enemy enjoy enough enter equal entrance escape even evening event ever every
everyone exact example except excited exercise expect explain eye face fact fail
fall false family famous far farm father fast fat fault fear feed feel
fever few fight fill film find fine finger finish fire first fish fit five fix
flag flat float floor flour flower fly fold food fool foot football for force
foreign forest forget forgive fork form fox four free freedom freeze fresh
friend friendly from front fruit full fun funny further future game garden gate
general get gift give glad glass go goat god gold good goodbye grass grave great
green gray ground group grow gun hair half hall hammer hand happen happy hard
hat hate have he head healthy hear heavy heart heaven height hello help hen her
here hers hide high hill him his hit hobby hold hole holiday home hope horse
hospital hot hotel house how hundred hungry hour hurry husband hurt ice idea if
in increase inside into invent iron invite is island it its jelly job join juice
jump just keep key kill kind king kitchen knee knife knock know ladder lady lamp
land large last late lately laugh lazy lead leaf learn leave leg left lend
length less lesson let letter library lie life light like lion lip list listen
little live lock lonely long look lose lot love low lower luck obey object ocean
of off offer office often oil old on one only open opposite or orange order
other our out outside over own page pain paint pair pan paper parent park part
partner party pass past path pay peace pen pencil people pepper per perfect
period person petrol piano pick picture piece pig pin pink place plane plant
plastic plate play please pleased plenty pocket point poison police polite pool
poor popular position possible potato pour power present press pretty prevent
price prince prison private prize probably problem produce promise proper
protect provide public pull punish pupil push put queen question quick quiet
quite radio rain rainy raise reach read ready real really receive record red
remember remind remove rent repair repeat reply report rest result return rice
rich ride right ring rise road rob rock room round rubber rude rule ruler run
rush sad safe sail salt same sand save say school science scissors search seat
second see seem sell send sentence serve seven several shade shadow shake
shape share sharp she sheep sheet shelf shine ship shirt shoe shoot shop short
should shoulder shout show sick side signal silence silly silver similar simple
single since sing sink sister sit six size skill skin skirt sky sleep slip slow
small smell smile smoke snow so soap sock soft some someone son soon sorry sound
soup south space speak special speed spell spend spoon sport spread spring
square stamp stand star start station stay steal steam step still stomach stone
stop store storm story strange street strong student study stupid subject such
sudden sugar suitable summer sun sunny support sure surprise sweet swim sword
table take talk tall taste taxi tea teach team tear tell ten tennis terrible
test than that the their then there these thick thin thing think third this
though threat three tidy tie title to today toe together tomorrow tonight too
tool tooth top total touch town train tram travel tree trouble true trust twice
try turn type ugly uncle under unit until up use useful usual usually very
village voice visit wait wake walk want warm was wash waste watch water way we
weak wear weather wedding week weight welcome were well west wet what wheel when
where which while white who why wide wife wild will win wind window wine winter
wire wise wish with without woman wonder word work world worry yard yell yet you
young your zero zoo
"""
words = words.split()

names = """aiden azami basia bob caris carol dave duska erin evander fatima frank greg gwen haris heidi idris isla john judy kali kamal layla lucas mary malik nina noor olivia oscar peter peggy quentin quinci raisa rupert samir sybil talia trent ursula usher vanna virgil walter wendy xavier xena yuri yvonne zahara zeke"""
names = names.split()


def simple_password():
    """Creates a new simple password of the form word, number, word.

    Returns:
        str -- Password.
    """
    return secrets.choice(words) + str(secrets.randbelow(100)) + secrets.choice(words)


def make_random_user_list(number=None):
    """Makes a list of random users.

    Keyword Arguments:
        number {int} -- Number of names in random user list (if the number is too large it is truncated). If None is given it is selected randomly. (default: {None})

    Returns:
        list -- List of [user, pasword] tuples.
    """
    if number is None or number > len(names):
        nlist = names
    else:
        nlist = sample(names, number)

    lst = []
    for n in sorted(nlist):
        p = simple_password()
        lst.append([n, p])
    return lst


def make_numbered_user_list(number=None):
    """Makes a list of numbered users (rather than named users).

    Keyword Arguments:
        number {int} -- Number of users to include in the list, if None is given users equal to the number of names is returned. (default: {None})

    Returns:
        list -- List of [user, pasword] tuples (user is a number).
    """
    if number is None:
        number = len(names)
    lst = []
    for i in range(0, number):
        n = "user{}".format(i)
        p = simple_password()
        lst.append([n, p])
    return lst



#!/usr/bin/env python3

# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020-2021 Colin B. Macdonald
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Plom script to start a demo server for homework submissions.

Instructions:
  * Make a new directory
  * Run this script inside it
  * In a new terminal, run the Plom Client and connect to localhost.
"""

__copyright__ = "Copyright (C) 2020-2021 Andrew Rechnitzer, Colin B. Macdonald et al"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import os
import subprocess
from shlex import split
import time
import argparse
from warnings import warn

from plom import __version__


parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)


def main():
    args = parser.parse_args()
    print("Plom version {}".format(__version__))

    if len(os.listdir(os.getcwd())) != 0:
        print('We recommend calling "{}" in an empty folder!'.format(parser.prog))
        warn("Current directory not empty")
    for f in (
        "specAndDatabase",
        "serverConfiguration",
        "archivedPDFs",
        "pageImages",
        "scanPNGs",
        "pages",
    ):
        if os.path.exists(f):
            raise RuntimeError('Directory "{}" must not exist for this demo.'.format(f))

    subprocess.check_call(split("plom-build new --demo"))
    subprocess.check_call(split("plom-server init"))
    subprocess.check_call(split("plom-server users --demo"))

    # Start server into background
    print("Running server I hope")
    serverproc = subprocess.Popen(split("plom-server launch"))
    time.sleep(1.0)
    try:
        serverproc.wait(1.0)
    except subprocess.TimeoutExpired:
        pass
    else:
        r = serverproc.returncode
        print("Server has prematurely stopped with return code {}".format(r))
        # TODO: server could send specific return code for "address already in use"?
        msg = "Server didn't start.  Is one already running?  See errors above."
        # raise RuntimeError(msg) from None
        print(msg)
        exit(r)

    assert serverproc.returncode is None, "has the server died?"

    print("Server seems to be running, so we move on to uploading")

    subprocess.check_call(split("plom-build class --demo -w 1234"))
    subprocess.check_call(split("plom-build make -w 1234"))
    # this creates two batches of fake hw - prefixes = hwA and hwB
    subprocess.check_call(split("plom-fake-hwscribbles -w 1234"))

    print("Processing some individually")
    # TODO: this is fragile, should not hardcode these student numbers!
    subprocess.check_call(
        split(
            "plom-hwscan process submittedHWByQ/semiloose.11015491._.pdf 11015491 -q 1,2,3 -w 4567"
        )
    )
    subprocess.check_call(
        split(
            "plom-hwscan process submittedHWByQ/semiloose.11135153._.pdf 11135153 -q 1,2,3 -w 4567"
        )
    )

    print("Processing all hw by question submissions.")
    subprocess.check_call(split("plom-hwscan allbyq -w 4567 -y"))
    print("Replacing all missing questions.")
    subprocess.check_call(split("plom-hwscan missing -w 4567 -y"))
    # print(">> TODO << process loose pages")

    time.sleep(0.5)
    try:
        serverproc.wait(0.5)
    except subprocess.TimeoutExpired:
        pass
    else:
        r = serverproc.returncode
        print("Server has prematurely stopped with return code {}".format(r))
        msg = "Server may have unexpectedly died during uploading.  See errors above."
        print(msg)
        exit(r)

    print('\n*** Now run "plom-client" ***\n')
    # TODO: output account info directly, perhaps just "user*"?
    print('  * See "serverConfiguration/userListRaw.csv" for account info\n')
    print("  * Press Ctrl-C to stop this demo")
    serverproc.wait()


if __name__ == "__main__":
    main()



html = """
<!DOCTYPE html>
<!-- Copyright (C) 2018-2020 Colin B. Macdonald <cbm@m.fsf.org>-->
<!-- SPDX-License-Identifier: FSFAP -->
<!--
     Copying and distribution of this file, with or without modification,
     are permitted in any medium without royalty provided the copyright
     notice and this notice are preserved.  This file is offered as-is,
     without any warranty.
-->
<html lang="en" dir="ltr" class="client-nojs">
<head>
<title>__COURSENAME__ - Online Return</title>

<meta charset="UTF-8" />
<style>
body {
	padding: 2em;
}

label {
    font-weight: bold;
    font-style: italic;
	font-size: medium;
    width: 12em;
    display: inline-block;
}

button {
	font-weight: bold;
	margin-top: 1em;
	line-height: 2;
}

</style>

<script>
	function retrievePaper() {
		var id = document.getElementById('studentID').value;
		if( !id || id.length!=__SID_LENGTH__ ) {
			alert("Invalid student number!");
			return;
		}
		var code = document.getElementById('studentCode').value;
		code = code.replace(/,/g,'')
		if( !code || code.length!=__CODE_LENGTH__ ) {
			alert("Invalid code!");
			return;
		}
		var which = '__TESTNAME__'
		window.location.href = which+"_"+id+"_"+code+".pdf";
	}
</script>
</head>



<body>
<hr>
<h1>__COURSENAME__ - Online Return</h1>

<p>This form can be used to retrieve an electronic copy of your paper.</p>

<hr>

<p>In order to access your paper you need to fill in your student number and
your __CODE_LENGTH__-digit &ldquo;return code&rdquo; from <a href="http://canvas.ubc.ca">canvas</a>.</p>


<p><i>NOTE:</i> the request will fail if incorrect information is entered.<p>

<div>
	<label class="label" for="studentID">Student number: </label>
	<input id="studentID" name="studentID" type="text" maxlength="__SID_LENGTH__" />
</div>
<br>
<div>
	<label class="label" for="studentCode">Code: </label>
	<input id="studentCode" name="studentCode" type="text" />
</div>
<br>
<div>
	<label class="label" for="retrieve"></label>
	<button id="retrieve" name="retrieve" onclick="retrievePaper()" type="button">Retrieve paper</button>
</div>
<div>

<hr>
</body>
</html>
"""



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

"""Read in a Canvas-exported spreadsheet and add return codes.

`plom-finish` can create a webpage which students can access
via their Student ID and a private secret code.  This script
prepares a csv file for Canvas import containing the secret codes.
"""

__copyright__ = "Copyright (C) 2019-2020 Colin B. Macdonald and others"
__credits__ = ["The Plom Project Developers"]
__license__ = "AGPL-3.0-or-later"

import argparse

from plom.finish import canvas_csv_add_return_codes, canvas_csv_check_pdf

canvas_fromfile = "canvas_from_export.csv"
canvas_return_tofile = "canvas_return_codes_for_import.csv"
# TODO: check if former exists and latter does not, and give some
# basic instructions


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--salt",
        type=str,
        required=True,
        help="Per-course secret salt string (required). See docs for details.",
    )
    parser.add_argument(
        "--digits",
        type=int,
        default=9,
        metavar="N",
        action="store",
        help="Length of the secret code.  Defaults to 9.",
    )

    args = parser.parse_args()
    print('Salt string is "{}"'.format(args.salt))
    print('Number of digits is "{}"'.format(args.digits))
    print(
        """
    *** Warning: this script is "pre-alpha" software ***

    This script looks for "{0}", which you should
    have exported from Canvas.  It outputs a new .csv files for
    importing back into canvas:

      * "{1}":
        The "return code" column will be filled.  Any existing
        return codes will be checked to confirm correctness.

    Read "docs/returning_papers.md" before using this.
    """.format(
            canvas_fromfile, canvas_return_tofile
        )
    )
    input("Press Enter to continue...")

    print()
    sns = canvas_csv_add_return_codes(
        canvas_fromfile, canvas_return_tofile, saltstr=args.salt, digits=args.digits
    )

    print()
    canvas_csv_check_pdf(sns)
